from myrrh.framework.msh.madvsh import AbcAdvSh

_cmd_internal = (
    b"true",
    b".",
    b"alias",
    b"bg",
    b"command",
    b"cd",
    b"echo",
    b"eval",
    b"exec",
    b"exit",
    b"export",
    b"fc",
    b"fg",
    b"getopts",
    b"hash",
    b"pwd",
    b"read",
    b"readonly",
    b"printf",
    b"set",
    b"shift",
    b"test",
    b"times",
    b"trap",
    b"type",
    b"ulimit",
    b"umask",
    b"unalias",
    b"unset",
    b"$",
    b"!",
    b"elif",
    b"fi",
    b"while",
    b"case",
    b"else",
    b"for",
    b"then",
    b"{",
    b"}",
    b"do",
    b"done",
    b"until",
    b"if",
    b"esac",
)

_cmd_operators = set((b"&&", b"|", b"||", b";", b";;", b"&", b">>", b"<<", b"<", b">", b">|", b">&", b"<&", b"<<-", b"<>", b"(", b")"))


__mlib__ = "AdvSh"


class AdvSh(AbcAdvSh):
    def initialize_stat(self):
        self.execute(b"%(echo)s init" % self.myrrh_os.getbinb, count=1).values()

    def _calibration_execution(self, count):
        return self.execute(b"%(sleep)s 1" % self.myrrh_os.getbinb, count=count)



import ntpath

from myrrh.utils import mstring
from myrrh.framework.mfs.madvfs import AbcAdvFs
from myrrh.core.services.system import (
    ExecutionFailureCauseRVal,
    ExecutionFailureCauseErr,
)

__mlib__ = "AdvFs"


class AdvFs(AbcAdvFs):
    default_mode = ""

    CHUNK_HEADER_SCRIPT = b"""\
@echo off
set vbscript=%%~F0.vbs
> %%vbscript%% echo pos=0:path="%%~F0"
>> %%vbscript%% echo Dim tStream : Set tStream = CreateObject("ADODB.Stream") : tStream.Open : tStream.LoadFromFile(path) : tStream.LineSeparator = 10 : tStream.CharSet = "utf-8"
>> %%vbscript%% echo do : line = tStream.ReadText(-2) : loop until tStream.EOS or line = "exit /b 0"
>> %%vbscript%% echo Dim s : line = tStream.ReadText(-2) : s = Split(line, ";")
>> %%vbscript%% echo Dim f : line = tStream.ReadText(-2) : f = Split(line, ";")
>> %%vbscript%% echo Dim iStream : Set iStream = CreateObject("ADODB.Stream") : iStream.Type = 1 : iStream.Open : iStream.LoadFromFile(path) : iStream.Position = tStream.Position-2
>> %%vbscript%% echo tStream.Close
>> %%vbscript%% echo For i = 0 to UBound(s)
>> %%vbscript%% echo Dim oStream : Set oStream = CreateObject("ADODB.Stream") : oStream.Type = 1 : oStream.Open
>> %%vbscript%% echo iStream.CopyTo oStream , s(i)
>> %%vbscript%% echo oStream.SaveToFile f(i), 2
>> %%vbscript%% echo oStream.Close
>> %%vbscript%% echo Next
>> %%vbscript%% echo WScript.Quit
%(cscript)s %%vbscript%%
del %%vbscript%%
(goto) 2>nul & del %%~F0 & exit /b %%errorlevel%%
exit /b 0"""

    def chown(self, path, user, group):
        # _, _, rval = self.myrrh_os.shell.execute(['/bin/chown %s:%s %s' % (user, group, path)])
        return True

    def chmod(self, mode, path):
        # _, _, rval = self.myrrh_os.shell.execute(['/bin/chmod %s %s' % (mode, path)])
        return True

    def _localtransfer(self, srcs, dest):
        if len(srcs) > 1:
            # TODO: test and refactoring required
            srcs = b" ".join((b'"%s"' % self.myrrh_os.sh_escape_bytes(s) for s in srcs))
            _, err, rval = self.myrrh_os.cmdb(
                b'if EXIST "%(dest)s\\" ( for %%f in (%(srcs)s) do if EXIST %%f\\ (%(robocopy)s /SL /E /NP %%f %%f) else (%(copy)s /Y /L %%f "%(dest)s" || exit 10) ) else exit 10',
                srcs=srcs,
                dest=self.myrrh_os.sh_escape_bytes(dest),
            )

        else:
            src = srcs[0]
            src = src.replace(b"/", self.myrrh_os._sepb_).replace(b"%", b"%%")
            dest = dest.replace(b"/", self.myrrh_os._sepb_).replace(b"%", b"%%")
            _, err, rval = self.myrrh_os.cmdb(
                b'if NOT EXIST "%(srcs)s\\" ( %(copy)s /Y /L "%(srcs)s" "%(dest)s" || exit 10 ) else if EXIST "%(dest)s\\" ( %(robocopy)s /E /SL /NP "%(srcs)s" "%(dest)s\\%(basename)s" ) else if EXIST "%(dirname)s" ( %(robocopy)s /E /SL /NP "%(srcs)s" "%(dest)s" ) else exit 10',
                srcs=self.myrrh_os.sh_escape_bytes(src),
                basename=self.myrrh_os.sh_escape_bytes(ntpath.basename(src)),
                dirname=self.myrrh_os.sh_escape_bytes(ntpath.dirname(dest) or src),
                dest=self.myrrh_os.sh_escape_bytes(dest),
            )

        err = b"No such file or directory" if rval == 10 else (b"Not a directory" if rval == 16 else err)
        rval = 0 if rval < 10 else 1

        ExecutionFailureCauseRVal(self, err, rval, 0).check()

    def _localmove(self, srcs, dest):
        if len(srcs) > 1:
            srcs = b" ".join((b'"%s"' % self.myrrh_os.sh_escape_bytes(s) for s in srcs))
            _, err, rval = self.myrrh_os.cmdb(
                b'if NOT EXIST "%(dest)s\\" (exit 10) else for %%f in (%(srcs)s) do %(move)s %%f "%(dest)s"',
                srcs=srcs,
                dest=self.myrrh_os.sh_escape_bytes(dest),
            )
        else:
            _, err, rval = self.myrrh_os.cmdb(
                b'%(move)s "%(srcs)s" "%(dest)s"',
                srcs=self.myrrh_os.sh_escape_bytes(srcs[0]),
                dest=self.myrrh_os.sh_escape_bytes(dest),
            )

        ExecutionFailureCauseRVal(self, err, rval, 0).check()

    def _localremove(self, srcs):
        srcs = b" ".join((b'"%s"' % self.myrrh_os.sh_escape_bytes(s) for s in srcs))
        _, err, rval = self.myrrh_os.cmdb(
            b"for %%f in (%(srcs)s) do if EXIST %%f\\ ( %(rmdir)s /Q /S %%f ) else del /F /Q %%f",
            srcs=srcs,
        )

        ExecutionFailureCauseRVal(self, err, rval, 0).check()

    def _chunk_header(self, file_descs):
        szs = b";".join((str(f[2]).encode() for f in file_descs))
        paths = b";".join((self.myrrh_os.p(f[1]) for f in file_descs))
        script = b"\n".join((self.CHUNK_HEADER_SCRIPT % {b"cscript": self.myrrh_os.getbinb[b"cscript"]}).splitlines())
        script = b"\n".join((script, szs, paths)) + b"\n"
        return script

    def _unchunk(self, path):
        path = self.myrrh_os.p(path)
        path_bat = b".".join((path, b"bat"))
        _, e, r = self.myrrh_os.cmdb(
            b'%(move)s "%(path)s" "%(path_bat)s" && "%(path_bat)s"',
            path=self.myrrh_os.sh_escape_bytes(path),
            path_bat=self.myrrh_os.sh_escape_bytes(path_bat),
        )
        ExecutionFailureCauseErr(self, e, b"").check()

    def _write_chunk(self, path, file_descs):
        path = self.myrrh_os.p(path)
        files = b"+".join((b'"%s"' % self.myrrh_os.sh_escape_bytes(self.myrrh_os.p(file)) for file, _, _ in file_descs))
        _, e, r = self.myrrh_os.cmdb(
            b'%(copy)s /B /Y %(files)s "%(path)s"',
            files=files,
            path=self.myrrh_os.sh_escape_bytes(path),
        )
        ExecutionFailureCauseRVal(self, e, r, 0).check()

    def _mkdirs(self, dirs: list):
        cmd = b"for %%d in ( %(dirs)s ) do if NOT EXIST %%d %(mkdir)s %%d"

        dirs = b" ".join((b'"%s\\"' % self.myrrh_os.sh_escape_bytes(d) for d in dirs))

        _, e, r = self.myrrh_os.cmdb(cmd, dirs=dirs)
        ExecutionFailureCauseRVal(self, e, r, 0).check()

    def _scanfiles(self, files):
        """
        return files(path, size)
        """
        files = b" ".join((b'"%s"' % self.myrrh_os.sh_escape_bytes(f) for f in files))
        o, e, r = self.myrrh_os.cmdb(b"for %%f in ( %(files)s ) do @echo %%f;%%~zf", files=files)
        ExecutionFailureCauseRVal(self, e, r, 0).check()

        output = (o.split(b";") for o in o.splitlines())

        files = []
        for path, sz in output:
            files.append(
                (
                    self.myrrh_os.shdecode(path).encode().strip(b'"'),
                    mstring.str2intb(sz),
                )
            )

        return files

    def _scantree(self, path):
        path = self.myrrh_os.p(path)

        o, e, r = self.myrrh_os.cmdb(
            b'for /f "usebackq delims=="  %%f in ( `dir /B /s` ) do @echo %%f;%%~af;%%~zf',
            execute_working_dir=path,
        )
        ExecutionFailureCauseRVal(self, e, r, 0).check()

        output = (o.split(b";") for o in o.splitlines())

        dirs = []
        files = []
        for path, ty, sz in output:
            if ty and ty[0] == ord(b"d"):
                dirs.append(self.myrrh_os.shdecode(path).encode())
            else:
                files.append((self.myrrh_os.shdecode(path).encode(), mstring.str2intb(sz)))

        return dirs, files



import typing

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from . import mbuiltins

__mlib__ = "AbcTempfile"


class _interface(ABC):
    import tempfile as local_tempfile

    @property
    @abstractmethod
    def NamedTemporaryFile(self):
        return self.local_tempfile.NamedTemporaryFile

    @property
    @abstractmethod
    def TemporaryFile(self):
        return self.local_tempfile.TemporaryFile

    @property
    @abstractmethod
    def SpooledTemporaryFile(self) -> local_tempfile.SpooledTemporaryFile:
        ...

    @property
    @abstractmethod
    def TemporaryDirectory(self) -> local_tempfile.TemporaryDirectory:
        ...

    @property
    @abstractmethod
    def TMP_MAX(self):
        return self.local_tempfile.TMP_MAX

    @property
    @abstractmethod
    def template(self) -> str:
        ...

    @property
    @abstractmethod
    def tempdir(self) -> str | None:
        ...

    @property
    @abstractmethod
    def _text_openflags(self) -> int:
        ...

    @property
    @abstractmethod
    def _bin_openflags(self) -> int:
        ...

    @property
    @abstractmethod
    def _RandomNameSequence(self):
        return self.local_tempfile._RandomNameSequence

    @property
    @abstractmethod
    def _TemporaryFileWrapper(self) -> local_tempfile._TemporaryFileWrapper:
        ...

    @abstractmethod
    def _candidate_tempdir_list(self) -> list:
        ...

    @abstractmethod
    def _get_candidate_names(self) -> list:
        ...

    @abstractmethod
    def _get_default_tempdir(self) -> typing.Any:
        ...

    @abstractmethod
    def _infer_return_type(*args) -> typing.Any:
        ...

    @abstractmethod
    def _mkstemp_inner(dir, pre, suf, flags, output_type) -> tuple[int, str]:
        ...

    @abstractmethod
    def mkstemp(self, suffix=None, prefix=None, dir=None, text=False) -> tuple[int, str]:
        ...

    @abstractmethod
    def mkdtemp(self, suffix=None, prefix=None, dir=None) -> str:
        ...

    @abstractmethod
    def mktemp(self, suffix="", prefix=template, dir=None) -> str:
        ...

    @abstractmethod
    def gettempprefix(self) -> bytes:
        ...

    @abstractmethod
    def gettempdir(self) -> str:
        ...

    @abstractmethod
    def gettempprefixb(self) -> bytes:
        ...

    @abstractmethod
    def gettempdirb(self) -> bytes:
        ...


class AbcTempfile(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mtempfile"

    __all__ = _interface.local_tempfile.__all__

    __delegated__ = {_interface: _interface.local_tempfile}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_tempfile, self)

        self.__delegate__(_interface, mod)



import myrrh.core.services

__name__ = myrrh.core.services.__distname__.upper() + " Development Framework"

__doc__ = """
    current version: %(version)s

%(summary)s

%(full_description)s

License
=======

%(license)s

""" % {
    "summary": myrrh.core.services.cfg_get(section="@metadata@", key="Summary"),
    "version": myrrh.core.services.__version__,
    "full_description": myrrh.core.services.cfg_get(section="@metadata@", key="Description"),
    "license": myrrh.core.services.__license__,
}

__docformat__ = "restructuredtext"



from myrrh.core.services.system import ExecutionFailureCauseRVal

from myrrh.framework.mpython.msys import AbcSys

__mlib__ = "Sys"


class Sys(AbcSys):
    __executable: bytes | None = None
    _abiflags = ""

    @property
    def _executable(self):
        if not self.__executable:
            for exe in (
                b"python3.7",
                b"python3.6",
                b"python3.5",
                b"python3.4",
                b"python3",
                b"python",
            ):
                self.__executable, err, rval = self.myrrh_os.cmdb(b"%(which)s %(exe)s", exe=exe)
                if rval == 0:
                    break
            else:
                ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return self.__executable

    @property
    def _platform(self):
        return "linux"

    @property
    def abiflags(self):
        return self._get_sys_attr("abiflags")



from asyncio.events import get_event_loop
import os
import asyncio
import time
import logging
import threading

from prompt_toolkit import history, shortcuts, auto_suggest, document

from prompt_toolkit.application.current import get_app, get_app_session
from prompt_toolkit.filters import Condition, is_done
from prompt_toolkit.layout import (
    ConditionalContainer,
    Window,
    FormattedTextControl,
    Float,
)
from prompt_toolkit.layout.containers import FloatContainer, HSplit, VSplit, WindowAlign
from prompt_toolkit.key_binding import merge_key_bindings
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.patch_stdout import patch_stdout

import bmy

from myrrh.core.services import cfg_get
from myrrh.core.services.logging import log

from myrrh.utils import mshlex

from . import completion
from . import cmd
from . import bindings
from . import exceptions

from traceback import print_exc
from prompt_toolkit.auto_suggest import Suggestion

HISTORY_CHAR = "!"


class History(history.FileHistory):
    def __init__(self) -> None:
        super().__init__(os.path.expanduser(os.path.join(cfg_get("@etc@", default=["~"])[0], ".myrrhc.history")))


class AutoSuggest(auto_suggest.AutoSuggest):
    def __init__(self):
        super().__init__()
        self.auto_suggest = auto_suggest.AutoSuggestFromHistory()

    def get_suggestion(self, buffer, document):
        history = buffer.history

        text = document.text.rsplit("\n", 1)[-1]

        if text.startswith("!"):
            text = text[1:]

            for string in reversed(list(history.get_strings())):
                for line in reversed(string.splitlines()):
                    if line.startswith(text):
                        return auto_suggest.Suggestion(line[len(text) :])


def create_exit_confirmation(prompt, style="class:exit-confirmation"):
    def get_text_fragments():
        # Show "Do you really want to exit?"
        return [
            (style, "\n %s ([y]/n) " % prompt.exit_message),
            ("[SetCursorPosition]", ""),
            (style, "  \n"),
        ]

    visible = ~is_done & Condition(lambda: prompt.show_exit_confirmation)

    return ConditionalContainer(
        content=Window(FormattedTextControl(get_text_fragments, focusable=True), style=style),
        filter=visible,
    )


class PromptSession(shortcuts.PromptSession):
    TITLE = "myrrhc"
    INFO = """
 Welcome to Myrrhc Console (for test purpose only)
  - Use it with fairplay -
Type help, copyright, credits, license for more information

ctrl-D to exit console

"""
    COMPLETER = completion.click_completer

    def __init__(self, ctx):
        # message cache
        self.__message_cache = None
        self.__info_task = None

        self._group_ctx = ctx.parent or ctx
        self._group = self._group_ctx.command
        self._group.options_metavar = ""
        self._group.set_ident(threading.get_ident())

        repl_command_name = ctx.command.name
        if isinstance(self._group_ctx.command, cmd.CommandCollection):
            available_commands = {cmd_name: cmd_obj for source in self._group_ctx.command.sources for cmd_name, cmd_obj in source.commands.items()}
        else:
            available_commands = self._group_ctx.command.commands

        available_commands.pop(repl_command_name, None)

        # Currently show 'Do you really want to exit?'
        self.exit_message = "Do you really want to exit"
        self.show_exit_confirmation = False
        self.confirm_exit = True

        try:
            session = get_app_session()
            output = session.output
            shortcuts.set_title(self.TITLE)
        except Exception:
            from prompt_toolkit.output import DummyOutput

            output = DummyOutput()

        super().__init__(
            message=self._message_no_info,
            completer=self.COMPLETER(self._group),
            history=History(),
            enable_history_search=True,
            complete_style=shortcuts.CompleteStyle.MULTI_COLUMN,
            complete_in_thread=True,
            auto_suggest=AutoSuggest(),
            output=output,
        )

    def _create_layout(self):
        layout = super()._create_layout()

        bottom_toolbar = ConditionalContainer(
            VSplit(
                [
                    Window(
                        FormattedTextControl(
                            lambda: self._bottom_toolbar()[0],
                            style="class:bottom-toolbar.text",
                        ),
                        style="class:bottom-toolbar",
                        dont_extend_height=True,
                        height=Dimension(min=1),
                        align=WindowAlign.LEFT,
                    ),
                    Window(
                        FormattedTextControl(
                            lambda: self._bottom_toolbar()[1],
                            style="class:bottom-toolbar.text",
                        ),
                        style="class:bottom-toolbar",
                        dont_extend_height=True,
                        height=Dimension(min=1),
                        align=WindowAlign.CENTER,
                    ),
                    Window(
                        FormattedTextControl(
                            lambda: self._bottom_toolbar()[2],
                            style="class:bottom-toolbar.text",
                        ),
                        style="class:bottom-toolbar",
                        dont_extend_height=True,
                        height=Dimension(min=1),
                        align=WindowAlign.RIGHT,
                    ),
                ]
            ),
            filter=~is_done & Condition(lambda: get_app().renderer.height_is_known) & Condition(lambda: self._bottom_toolbar is not None),
        )

        self.exit_confirmation = create_exit_confirmation(self)

        new_layout = FloatContainer(
            content=HSplit([layout.container, bottom_toolbar]),
            floats=[
                Float(
                    left=2,
                    bottom=1,
                    content=self.exit_confirmation,
                )
            ],
        )

        return Layout(new_layout, layout.container)

    def _create_prompt_bindings(self):
        return merge_key_bindings(
            [
                super()._create_prompt_bindings(),
                bindings.load_prompt_bindings(self),
                bindings.load_confirm_exit_bindings(self),
            ]
        )

    def _bottom_toolbar(self):
        if not cmd.option_cfg_default("fetch_eid_info", True)():
            if self.__info_task:
                self.__info_task.cancel()
                self.__info_task = None

            self.__message_cache == ("info disabled", "", "")

        elif self.__info_task is None:
            self._info_task()

        return self.__message_cache

    def _info_task(self):
        def _get_pwd(ident, cancel_event):
            while not cancel_event.is_set():
                try:
                    eid = bmy.current(ident=ident)
                    if not bmy.isgroup(eid):
                        l_info = cmd.option_cfg_default("fetch_eid_info_left", "cwd")().split(",")
                        m_info = cmd.option_cfg_default("fetch_eid_info_middle", "id")().split(",")
                        r_info = cmd.option_cfg_default("fetch_eid_info_right", "location")().split(",")
                    else:
                        l_info = cmd.option_cfg_default("fetch_eids_info_left", "cwd")().split(",")
                        m_info = cmd.option_cfg_default("fetch_eids_info_middle", "id")().split(",")
                        r_info = cmd.option_cfg_default("fetch_eids_info_right", "")().split(",")

                    if not eid or not cmd.option_cfg_default("fetch_eid_info", True)() or not bmy.isbuilt(eid=eid):
                        l_text = os.getcwd()
                        m_text = "*"
                        r_text = "*"
                    else:
                        info = bmy.info(
                            [
                                *(k.strip() for k in l_info),
                                *(k.strip() for k in m_info),
                                *(k.strip() for k in r_info),
                            ],
                            eid=eid,
                        )
                        if not bmy.isgroup(eid):
                            l_text = "|".join((str(info[k.strip()] or "na") for k in l_info)) if l_info else "*"
                            m_text = "|".join((str(info[k.strip()] or "na") for k in m_info)) if m_info else "*"
                            r_text = "|".join((str(info[k.strip()] or "na") for k in r_info)) if r_info else "*"
                        else:
                            l_text = "|".join(str(info._d_[i][k.strip()] or "na") for k in l_info for i in bmy.groupkeys(info)) if l_info else "*"
                            m_text = "|".join(str(info._d_[i][k.strip()] or "na") for k in m_info for i in bmy.groupkeys(info)) if m_info else "*"
                            r_text = "|".join(str(info._d_[i][k.strip()] or "na") for k in r_info for i in bmy.groupkeys(info)) if r_info else "*"

                except Exception as e:
                    l_text = "Unable to fetch information: %s" % str(e)
                    m_text = "*"
                    r_text = "*"

                self.__message_cache = (l_text, m_text, r_text)

                time.sleep(cmd.option_cfg_default("fetch_eid_info_delay", 1)())

        async def _coro():
            cancel_evt = asyncio.Event()

            try:
                await asyncio.to_thread(_get_pwd, threading.get_ident(), cancel_evt)
            finally:
                cancel_evt.set()
                self.__info_task = None

        self.__info_task = get_event_loop().create_task(_coro(), name="info")

    def _message_no_info(self):
        prompt = cmd.option_cfg_default("prompt", "")()
        if prompt:
            return prompt
        return cmd.format_eid()

    async def run(self):
        cmd.info(self.INFO)
        self.__message_cache = "..."

        try:
            while True:
                try:
                    with patch_stdout():
                        command = await self.prompt_async()
                    await self.command_call(command)

                except KeyboardInterrupt:
                    pass
        finally:
            if self.__info_task:
                self.__info_task.cancel()

    async def command_call(self, command):
        prev_status = 0
        args = self._parse(command)
        try:
            while True:
                arg = args.pop(0)

                if isinstance(arg, (tuple, list)):
                    with self._group.make_context(None, arg) as ctx:
                        try:
                            res = self._group.invoke(ctx)
                            if res:
                                await res()
                            prev_status = 0
                        except (exceptions.Reboot, exceptions.Abort, exceptions.Exit):
                            raise
                        except exceptions.Failure:
                            prev_status = 1
                        except Exception:
                            if log.level >= logging.DEBUG:
                                print_exc()
                            prev_status = 1

                elif arg == "&&" and prev_status != 0:
                    args.pop(0)

                elif arg == "||" and prev_status == 0:
                    args.pop(0)

        except IndexError:
            pass

    def _parse(self, command):
        if command and command.startswith(HISTORY_CHAR):
            suggestion = self.auto_suggest.get_suggestion(self.app.current_buffer, document.Document(text=command)) or Suggestion(text="")
            command = "".join([command[1:], suggestion.text])

            if not command.strip():
                commands = self.history.get_strings()
                command = commands and commands[-1] or ""

            self.history.append_string(command)

        if not command:
            return []

        args = mshlex.split(command)

        commands = []
        cmd = []
        for arg in args:
            if arg not in ("&&", "||", ";"):
                cmd.append(arg)
            else:
                commands.append(cmd)
                commands.append(arg)
                cmd = []

        commands.append(cmd)
        return commands



import os
import stat
from click.core import BaseCommand

import click.shell_completion
from prompt_toolkit import completion, eventloop

from myrrh.utils import mshlex
import bmy


class ThreadedCompleter:
    def get_completions_async(self, document, complete_event):
        """
        Asynchronous generator of completions.
        This yields both Future and Completion objects.
        """
        return eventloop.generator_to_async_generator(lambda: self.get_completions(document, complete_event) or ())


class _ShellComplete(click.shell_completion.ShellComplete):
    def get_completion_args(self):
        try:
            args = mshlex.split(self.complete_var.text_before_cursor)
        except ValueError:
            # Invalid command, perhaps caused by missing closing quotation.
            return

        current_args = []

        try:
            while True:
                arg = args.pop()
                if arg not in ("&&", "||", ";"):
                    current_args.append(arg)
                else:
                    break

        except IndexError:
            pass

        current_args.reverse()
        args = current_args

        cursor_within_command = self.complete_var.text_before_cursor.rstrip() == self.complete_var.text_before_cursor
        if args and cursor_within_command:
            incomplete = args[-1]
            args = args[:-1]
        else:
            incomplete = ""

        return args, incomplete


class _ClickCompleter(completion.Completer, ThreadedCompleter):
    def __init__(self, cli: BaseCommand):
        self.cli = cli

    def get_completions(self, document, complete_event):
        shell_complete = _ShellComplete(self.cli, self.cli.context_settings, "", document)
        args, incomplete = shell_complete.get_completion_args()
        sh_completions = shell_complete.get_completions(args, incomplete) or []
        words = [a.value for a in sh_completions]

        return completion.WordCompleter(words).get_completions(document, complete_event=complete_event)


class _BmyPathShellComplete:
    def __init__(
        self,
        only_directories=False,
        get_paths=None,
        file_filter=None,
        min_input_len=0,
        expanduser=False,
        local=False,
        use_argument_for_eid="eid",
        file_text=None,
    ):
        assert get_paths is None or callable(get_paths)
        assert file_filter is None or callable(file_filter)
        assert isinstance(min_input_len, int)
        assert isinstance(expanduser, bool)

        self.only_directories = only_directories
        self.get_paths = get_paths or (lambda _: ["."])
        self.file_filter = file_filter or (lambda _: True)
        self.min_input_len = min_input_len
        self.expanduser = expanduser
        self.local = local
        self.use_argument_for_eid = use_argument_for_eid
        self.doc = None
        self.file_text = file_text or self._file_text

    def _file_text(self, dirent, ctx):
        if dirent.is_dir():
            filename = dirent.name + self._os(ctx).sep
        else:
            filename = dirent.name

        return filename

    def __call__(self, ctx, cli, incomplet):
        # Complete only when we have at least the minimal input length,
        # otherwise, we can too many results and autocompletion will become too
        # heavy.
        paths = []

        if len(incomplet) < self.min_input_len:
            return
        text = incomplet
        try:
            # Do tilde expansion.
            if self.expanduser:
                text = self._os(ctx).path.expanduser(text)

            # Directories where to look.
            dirname = self._os(ctx).path.dirname(text)
            if dirname.startswith(self._os(ctx).path.curdir) or dirname.startswith(self._os(ctx).path.pardir) or self._os(ctx).path.isabs(dirname):
                directories = [dirname]
            elif dirname:
                directories = [self._os(ctx).path.dirname(self._os(ctx).path.join(p, text)) for p in self.get_paths(ctx)]
            else:
                directories = self.get_paths(ctx)

            # Start of current file.
            prefix = self._os(ctx).path.basename(text)

            directories = sorted(filter(None, directories))

            # Get all filenames.
            for directory in directories:
                try:
                    # Look for matches in this directory.
                    files = filter(
                        lambda dirent: (dirent.name.startswith(prefix) and ((not self.only_directories or dirent.is_dir()) and self.file_filter(dirent))),
                        self._os(ctx).scandir(directory),
                    )
                    files = sorted(files, key=lambda d: d.name)
                    for dirent in files:
                        filename = self.file_text(dirent, ctx)
                        filename = self._os(ctx).path.sep + filename if text and text[-1] == self._os(ctx).path.sep else filename
                        completion = click.shell_completion.CompletionItem(filename, help=self._os(ctx).path.dirname(dirent.path))
                        paths.append(completion)
                except Exception:
                    pass

        except Exception:
            pass

        return paths

    def _os(self, ctx):
        eid = None

        eid = ctx.params.get(self.use_argument_for_eid, None) or bmy.current(ident=ctx.ident)
        if not bmy.isgroup(eid) and eid:
            eid = (eid,)

        if not eid:
            return os

        if len(eid) == 1:
            try:
                with bmy.select(eid=eid[0]):
                    from mlib.py import os
                return os
            except Exception:
                pass

        raise OSError


class _LocalPathShellComplete(_BmyPathShellComplete):
    def _os(self, ctx):
        return os


class _BmyExecutableShellComplete(_BmyPathShellComplete):
    """
    Complete only executable files in the current path.
    """

    def __init__(self):
        super().__init__(
            only_directories=False,
            min_input_len=1,
            get_paths=lambda ctx: self._os(ctx).environ.get("PATH", "").split(self._os(ctx).pathsep) + [self._os(ctx).getcwd()],
            file_filter=lambda dirent: stat.S_IMODE(dirent.stat().st_mode) & (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH),
            expanduser=True,
        ),


class _LocalExecutableShellComplete(_BmyPathShellComplete):
    """
    Complete only executable files in the current local path.
    """

    def __init__(self):
        super().__init__(
            only_directories=False,
            min_input_len=1,
            get_paths=lambda ctx: self._os(ctx).environ.get("PATH", "").split(self._os(ctx).pathsep) + [self._os(ctx).getcwd()],
            file_filter=lambda ctx, dirent: self._os(ctx).access(dirent.path, os.X_OK),
            expanduser=True,
        )

    def _os(self, ctx):
        return os


def eids_completer(_ctx, _cli, _incomplet):
    return bmy.eids()


path_completer = _BmyPathShellComplete
path_executor = _BmyExecutableShellComplete
local_path_completer = _LocalPathShellComplete
local_path_executor = _LocalExecutableShellComplete

click_completer = _ClickCompleter



import time
import stat
import os
import re
import sys
import calendar

from myrrh.utils.mstring import str2int

from myrrh.core.services.system import (
    ExecutionFailureCauseRVal,
    ExecutionFailureCauseErr,
    _mlib_,
)
from myrrh.core.interfaces import ABCDelegation, ABC, abstractmethod

from myrrh.framework.mpython import mbuiltins
from myrrh.framework.mpython._mosfs import AbcOsFs, stat_result

__mlib__ = "OsFs"


class _interface(ABC):
    import ntpath as local_path

    @abstractmethod
    def normcase(self, s):
        ...

    @abstractmethod
    def splitdrive(self, p):
        ...

    @abstractmethod
    def split(self, p):
        ...

    @abstractmethod
    def expandvars(self, path):
        ...

    @abstractmethod
    def _get_bothseps(self, path):
        ...


class OsFs(_interface, AbcOsFs, ABCDelegation):
    _umask = 0o000
    _epochdelta = 0 if sys.version_info >= (3, 6) else 3600
    _PATHEXT = {b".COM", b".EXE", b".BAT", b".CMD", b".VBS", b".JS", b".WS", b".MSC"}

    __delegated__ = {_interface: _interface.local_path}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_path, self, {"os": self})

        self.__delegate__(_interface, mod)

    def ismount(self, mount):
        mount = self.myrrh_os.p(mount)
        return re.match(b"([a-zA-Z]:$)|(//.*)", mount) is not None

    def _wmidatetoiso(self, str):
        str = self.fsdecode(str)
        date = re.match(r"(\d+.\d+)([+|-])(\d+|\*+)", str)
        if date is None:
            date = re.match(r"(\d+.\d+)([+|-])(\d+|\*+)", "197603090000.000000+000")
        date = date.groups()
        tz = str2int(date[2], 0)
        return "%s%s%02d%02d" % (date[0], date[1], tz // 60, tz % 60)

    def chown(self, path, uid, gid, *, dir_fd=None, follow_symlinks=True, _trusted=False):
        if _trusted:
            return

        if not isinstance(uid, int):
            raise TypeError("invalid user type")
        if not isinstance(gid, int):
            raise TypeError("invalid group type")

        if uid == -1 and gid == -1:
            return

        self._stat(path)

    def chmod(self, path, mode, *, dir_fd=None, follow_symlinks=True, _trusted=False):
        if _trusted:
            return

        self._stat(path)

    def rename(self, src, dst, *, src_dir_fd=None, dst_dir_fd=None):
        _src = self.myrrh_os.p(src, dir_fd=src_dir_fd)
        _dst = self.myrrh_os.p(dst, dir_fd=dst_dir_fd)

        _src = self.abspath(_src)
        _dst = self.abspath(_dst)

        samepath = _src.lower() == _dst.lower()

        cmd, _dst = (b'%(rename)s "%(src)s" "%(dst)s"', self.myrrh_os.basename(_dst)) if samepath else (b'%(move)s /Y "%(src)s" "%(dst)s"', _dst)

        _, err, rval = self.myrrh_os.cmd(
            cmd,
            src=self.myrrh_os.sh_escape_bytes(_src),
            dst=self.myrrh_os.sh_escape_bytes(_dst),
        )

        ExecutionFailureCauseRVal(self, err, rval, 0, src).check()

    def listdir(self, path="."):
        _cast_ = self.myrrh_os.fdcast(path)
        path = self.myrrh_os.normpath(path)

        _path = self.myrrh_os.p(path)

        out, err, rval = self.myrrh_os.cmdb(
            b'if exist "%(path)s\\"  ( %(dir)s /B "%(path)s" ) else if exist "%(path)s" (exit 20) else ( exit 2 )',
            path=self.myrrh_os.sh_escape_bytes(_path),
        )

        ExecutionFailureCauseRVal(self, err, rval, 0, _cast_(path), errno=rval).check()
        return [_cast_(self.myrrh_os.shdecode(f)) for f in [f.strip() for f in out.split(self.myrrh_os.linesepb)] if f != self.myrrh_os.curdirb and f != self.myrrh_os.pardirb and len(f) != 0]

    def _scandir_list(self, path="."):
        # wmic output always in os encoding

        _cast_ = self.myrrh_os.fdcast(path)
        _path = self.myrrh_os.p(path)
        _abspath = self.abspath(_path)
        wmicpath = _abspath.replace(b"\\", b"\\\\").replace(b"'", b"\\'").replace(b'"', b'\\"')

        drive, wmicpath = self.splitdrive(wmicpath)
        wmicpath = b"" if wmicpath == b"\\\\" else wmicpath
        out, err, rval = self.myrrh_os.cmd(
            b"%(dir)s /B \"%(path)s\" && %(echo)s \"OOO___:__OOOO\" && %(wmic)s fsdir where \"path='%(wmicpath)s\\\\' and drive='%(drive)s'\" get %(property)s /value && %(wmic)s datafile where \"path='%(wmicpath)s\\\\' and drive='%(drive)s'\" get %(property)s /value",
            path=self.myrrh_os.sh_escape_bytes(_path),
            drive=drive,
            wmicpath=wmicpath,
            property=self._stat_property,
        )
        ExecutionFailureCauseRVal(self, err, rval, 0, path).check()

        # try translate to utf8 bytes
        try:
            out = out.encode()
            err = err.encode()
        except ValueError:
            pass

        out_dir, out_wmi = out.split(b'"OOO___:__OOOO"', 1)
        real_file_name = {o.lower(): o for o in out_dir.split(b"\r\n")}

        # utf16 hack
        if b"\x00" in out_wmi:
            out_wmi = out_wmi.replace(b"\x00", b"")

        out_wmi = re.split(b"[\\r\\n]{4,}", out_wmi)

        result = []
        for stat_ in (s for s in out_wmi if s.strip()):
            fpath, stat_ = self._stat_out_to_struct(stat_)
            fname = self.myrrh_os.basename(fpath)
            # find real name
            for fn_lower, fn in real_file_name.items():
                if fname == fn_lower:
                    fname = fn
                    break
            else:
                fn_lower = ""

            if fn_lower:
                real_file_name.pop(fn_lower)

            fpath = self.myrrh_os.joinpath(_path, fname)
            result.append(self.DirEntry(_cast_(fname), _cast_(fpath), stat_, self.lstat))

        return result

    def isdir(self, path):
        path = self.myrrh_os.p(path)

        if len(path) == 0:
            return False

        return self.myrrh_os.fs.is_container(path)

    def isfile(self, path):
        path = self.myrrh_os.p(path)

        if len(path) == 0:
            return False

        return self.myrrh_os.fs.exist(path) and not self.myrrh_os.fs.is_container(path)

    @property
    def _pathext(self):
        return self._PATHEXT.union(p.upper() for p in (self.myrrh_os.getenvb().get(b"PATHEXT", b"").split(self.myrrh_os.pathsepb)))

    _stat_property = b"name,filesize,readable,writeable,filetype,creationdate,lastaccessed,lastmodified,archive,compressed,encrypted,hidden,system"

    def _stat_out_to_struct(self, out):
        if b"\x00" in out:
            out = out.lstrip().rstrip(b"\x00")
            out = self.myrrh_os.shencode(out.decode("utf-16-be"))

        out_stat = {a[0]: a[1].strip() for a in (line.split(b"=", 1) for line in out.split(b"\r\n")) if len(a) == 2}

        mode = stat.S_IFDIR if out_stat.get(b"FileType", b"") in (b"File Folder", b"Local Disk") else stat.S_IFREG
        path = out_stat.get(b"Name", b"myrrh_unexpected_error???????")
        # read GR
        if out_stat.get(b"Readable", b"TRUE") == b"TRUE":
            access_mask_win = stat.S_IRUSR
        if out_stat.get(b"Writeable", b"TRUE") == b"TRUE":
            access_mask_win += stat.S_IWUSR
        _, ext = self.splitext(path)
        if ext.upper() in self._pathext:
            access_mask_win += stat.S_IXUSR

        mode += (stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH) if (access_mask_win & stat.S_IRUSR) == stat.S_IRUSR or (access_mask_win & stat.S_IFDIR) == stat.S_IFDIR else 0
        mode += (stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH) if (access_mask_win & stat.S_IWUSR) == stat.S_IWUSR or (access_mask_win & stat.S_IFDIR) == stat.S_IFDIR else 0
        mode += (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH) if (access_mask_win & stat.S_IXUSR) == stat.S_IXUSR or (mode & stat.S_IFDIR) == stat.S_IFDIR else 0

        file_size = out_stat.get(b"FileSize", b"")
        if file_size == b"":
            file_size = b"0"
        creation_date = out_stat.get(b"CreationDate", b"")
        if creation_date == b"":
            creation_date = b"19700101010000.000000+000"
        last_accessed = out_stat.get(b"LastAccessed", b"")
        if last_accessed == b"":
            last_accessed = creation_date
        last_modified = out_stat.get(b"LastModified", b"")
        if last_modified == b"":
            last_modified = creation_date

        def mktime(date):
            date = time.strptime(self._wmidatetoiso(date), "%Y%m%d%H%M%S.%f%z")
            date = calendar.timegm(date) - self._epochdelta
            return date, date * 1000000000

        creation_date, creation_date_ns = mktime(creation_date)
        last_accessed, last_accessed_ns = mktime(last_accessed)
        last_modified, last_modified_ns = mktime(last_modified)

        st_file_attributes = (
            (stat.FILE_ATTRIBUTE_ARCHIVE if out_stat.get(b"Archive", b"FALSE") == b"TRUE" else 0)
            | (stat.FILE_ATTRIBUTE_COMPRESSED if out_stat.get(b"Compressed", b"FALSE") == b"TRUE" else 0)
            | (stat.FILE_ATTRIBUTE_ENCRYPTED if out_stat.get(b"Encrypted", b"FALSE") == b"TRUE" else 0)
            | (stat.FILE_ATTRIBUTE_HIDDEN if out_stat.get(b"Hidden", b"FALSE") == b"TRUE" else 0)
            | (stat.FILE_ATTRIBUTE_READONLY if out_stat.get(b"Readable", b"FALSE") == b"TRUE" else 0)
            | (stat.FILE_ATTRIBUTE_SYSTEM if out_stat.get(b"System", b"FALSE") == b"TRUE" else 0)
            | (stat.FILE_ATTRIBUTE_DIRECTORY if out_stat.get(b"FileType", b"") in (b"File Folder", b"Local Disk") else stat.FILE_ATTRIBUTE_NORMAL)
        )

        result = stat_result(
            (
                mode,
                0,
                0,
                0,
                0,
                0,
                int(file_size, 10),
                last_accessed,
                last_modified,
                creation_date,
            ),
            {
                "st_file_attributes": st_file_attributes,
                "st_atime_ns": last_accessed_ns,
                "st_mtime_ns": last_modified_ns,
                "st_ctime_ns": creation_date_ns,
            },
        )

        return path, result

    # TODO: rewrite stat for optimization and correct file links management
    def _stat(self, path, *, dir_fd=None, follow_symlinks=True):
        _path = self.myrrh_os.f(path, dir_fd=dir_fd)
        _path = self.abspath(_path)

        if _path == b"\\\\.\\nul":
            return stat_result(8192, 0, 0, 0, 0, 0, 0, 0, 0, 0)

        if self.ismount(_path):
            out, err, rval = self.myrrh_os.cmdb(b'%(dir)s /o:d /b "%(path)s"', path=self.myrrh_os.sh_escape_bytes(_path))
            ExecutionFailureCauseRVal(self, err, rval, _path, 0).check()
            ExecutionFailureCauseErr(self, err, "").check()
            if out == "":
                return stat_result((0, 0, 0, 0, 0, 0, 0, 0, 0, 0), {})
            _path = self.myrrh_os.joinpath(_path, out.split(self.myrrh_os.sepb)[0])  # get first

        out, err, rval = self.myrrh_os.cmdb(
            b'if EXIST "%(path)s\\" ( %(wmic)s fsdir where "name=\'%(wmicpath)s\'" get %(property)s /value ) else if EXIST "%(path)s" ( %(wmic)s datafile where "name=\'%(wmicpath)s\'" get %(property)s /value ) else exit 2',
            path=self.myrrh_os.sh_escape_bytes(_path),
            wmicpath=self.myrrh_os.sh_escape_bytes(_path).replace(b"\\", b"\\\\").replace(b"'", b"\\'").replace(b'"', b'\\"'),
            property=self._stat_property,
        )

        ExecutionFailureCauseRVal(self, err, rval, 0, path, errno=rval).check()
        ExecutionFailureCauseErr(self, err, b"").check()

        _, result = self._stat_out_to_struct(out)
        return result

    def _lstat(self, path, *, dir_fd=None):
        return self.stat(path, dir_fd=dir_fd, follow_symlinks=False)

    def _statvfs(self, path):
        _path = self.myrrh_os.f(path)
        _path = self.abspath(_path)

        drive = self.splitdrive(_path)

        out, err, rval = self.myrrh_os.cmdb(b"%(wmic)s volume where driveletter='%(drive)s' get /value", drive=drive[0])
        ExecutionFailureCauseRVal(self, err, rval, 0, _path).check()
        ExecutionFailureCauseErr(self, err, b"").check()

        _stat = {a[0]: str2int(a[1], -1) for a in (line.split(b"=") for line in out.split(b"\r\n")) if len(a) == 2}

        block_sz = _stat[b"BlockSize"]
        blocks = _stat[b"Capacity"] / block_sz
        bfree = _stat[b"FreeSpace"] / block_sz
        namemax = _stat[b"MaximumFileNameLength"]

        return OsFs.statvfs_result(block_sz, block_sz, blocks, bfree, bfree, -1, -1, -1, -1, namemax)

    def realpath(self, path):
        path = self.myrrh_os.p(path)
        path = self.myrrh_os.normpath(path)

        dirname = self.myrrh_os.dirname(path)
        basename = self.myrrh_os.basename(path)

        out, err, rval = self.myrrh_os.cmd(
            b'%(cd)s /D "%(dirname)s" && FOR %%f in (%(basename)s) do @%(echo)s %%~ff ',
            dirname=b"." if len(dirname) == 0 else dirname,
            basename=basename,
        )
        ExecutionFailureCauseRVal(self, err, rval, 0, path).check()

        return out

    def expanduser(self, path):
        path = self.myrrh_os._fspath_(path)
        if isinstance(path, bytes):
            tilde = b"~"
        else:
            tilde = "~"
        if not path.startswith(tilde):
            return path
        i, n = 1, len(path)
        while i < n and path[i] not in self._get_bothseps(path):
            i += 1

        if "HOME" in os.environ:
            userhome = os.environ["HOME"]
        elif "USERPROFILE" in os.environ:
            userhome = os.environ["USERPROFILE"]
        elif "HOMEPATH" not in os.environ:
            return path
        else:
            try:
                drive = os.environ["HOMEDRIVE"]
            except KeyError:
                drive = ""
            userhome = self.myrrh_os.joinpath(drive, os.environ["HOMEPATH"])

        if isinstance(path, bytes):
            userhome = os.fsencode(userhome)

        if i != 1:  # ~user
            userhome = self.myrrh_os._joinpath__(self.myrrh_os.__m_dirname_(userhome), path[1:i])

        return userhome + path[i:]

    def replace(self, src, dst, *args, src_dir_fd=None, dst_dir_fd=None):
        self.rename(src, dst, *args, src_dir_fd=src_dir_fd, dst_dir_fd=dst_dir_fd)

    def _utime(self, path, times=None, *, ns=None, dir_fd=None, follow_symlinks=True):
        _path = self.myrrh_os.f(path, dir_fd=dir_fd)

        if times is None and ns is None:
            _, err, rval = self.myrrh_os.cmdb(b'%(copy)s /b "%(path)s"+,,', path=self.myrrh_os.sh_escape_bytes(_path))
        else:
            atime, mtime = times or (t * 1e-9 for t in ns)
            atime, mtime = time.gmtime(atime + self._epochdelta), time.gmtime(mtime + self._epochdelta)
            atime = b"%i/%i/%i %i:%i:%i" % (
                atime.tm_year,
                atime.tm_mon,
                atime.tm_mday,
                atime.tm_hour,
                atime.tm_min,
                atime.tm_sec,
            )
            mtime = b"%i/%i/%i %i:%i:%i" % (
                mtime.tm_year,
                mtime.tm_mon,
                mtime.tm_mday,
                mtime.tm_hour,
                mtime.tm_min,
                mtime.tm_sec,
            )

            _, err, rval = self.myrrh_os.cmdb(
                b"%(powershell)s set-itemproperty '%(path)s' -name LastAccessTime -Value '%(atime)s' ",
                atime=atime,
                path=self.myrrh_os.sh_escape_bytes(_path),
            )
            ExecutionFailureCauseRVal(self, err, rval, 0, path).check()
            _, err, rval = self.myrrh_os.cmdb(
                b"%(powershell)s set-itemproperty '%(path)s' -name LastWriteTime -Value '%(mtime)s' ",
                mtime=mtime,
                path=self.myrrh_os.sh_escape_bytes(_path),
            )

        ExecutionFailureCauseRVal(self, err, rval, 0, path).check()

    def link(self, src, dst, *, src_dir_fd=None, dst_dir_fd=None, follow_symlinks=True):
        _src = self.myrrh_os.p(src, dir_fd=src_dir_fd)
        _dst = self.myrrh_os.p(dst, dir_fd=dst_dir_fd)

        _, err, rval = self.myrrh_os.cmd(
            b'%(mklink)s "%(dst)s" "%(src)s"',
            dst=self.myrrh_os.sh_escape_bytes(_dst),
            src=self.myrrh_os.sh_escape_bytes(_src),
        )
        ExecutionFailureCauseRVal(self, err, rval, 0, src).check()

    def umask(self, mask):
        return 0o000 & mask

    def truncate(self, path, length):
        from myrrh.framework.mpython import _mosfile

        mf = _mlib_(_mosfile)(self)
        fd = mf.dup(path) if (isinstance(path, int)) else mf.open(path, mf.O_APPEND)
        try:
            mf.ftruncate(fd, length)
        finally:
            mf.close(fd)

    def samefile(self, path1, path2):
        return self.normcase(self.abspath(path1)) == self.normcase(self.abspath(path2))

    @property
    def environ(self):
        return self.myrrh_os.getenv()



import contextlib
import threading
import time
import typing

from myrrh.warehouse.item import NoneItem

from ...services import cfg_init

__all__ = ("Acquiring", "runtime_cached_property", "RuntimeCache", "init_cache")


class Acquiring(Exception):
    ...


class RuntimeCache(dict):
    def __init__(self):
        super().__init__(_runtime_cache)
        self.__lock__ = threading.RLock()
        self.__status__ = _new_runtime_status()

    @contextlib.contextmanager
    def attr(self, attr, default=None):
        with self.__lock__:
            yield self.get(attr, default)

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            pass

        raise AttributeError(attr)

    def __setattr__(self, attr, value):
        try:
            self[attr] = value
            return
        except KeyError:
            pass

        raise AttributeError(attr)


def init_cache(cache, system):
    for k, d in _runtime_status.items():
        if getattr(system.__class__, d["property"].property_name, None):
            cfg_path = d["init_cfg_path"]

            if cfg_path and system.cfg[cfg_path] is not NoneItem:
                cache[k] = system.cfg[cfg_path]

            if d["init_at_creation_time"]:
                cache[k] = d["property"].get(system, cache)

    system.__m_runtime_cache__ = cache


_runtime_cache = {"__lock__": None, "__status__": None}
_runtime_status: dict[str, dict[str, typing.Any]] = dict()


def _new_runtime_status():
    d = dict()
    for k, v in _runtime_status.items():
        d[k] = dict(v)

    return d


class _RuntimeProperty:
    ACQ = 2
    SET = 1
    UNSET = 0

    def __init__(self, func, name):
        self.name = name
        self.func = (lambda s: getattr(s, func)) if isinstance(func, str) else func
        self.__doc__ = func.__doc__

    def __set_name__(self, owner, name):
        self.property_name = name

    def get(self, instance, cache):
        # optim?
        status = cache.__status__[self.name]
        if not (status["validity"] == -1 and status["state"] == self.SET):
            with cache.__lock__:
                if status["state"] == self.ACQ:
                    raise Acquiring

                isvalid = status["state"] == self.SET and (status["validity"] == -1 or (time.monotonic() > (status["date"] + status["validity"])))

                if isvalid:
                    return cache[self.name]

                status["state"] = self.ACQ

                self.set(instance, cache, self.func(instance))

        return cache[self.name]

    def set(self, instance, cache, value):
        with cache.__lock__:
            cache[self.name] = value
            cache.__status__[self.name]["date"] = time.monotonic()
            cache.__status__[self.name]["state"] = self.SET

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return self.get(instance, instance.__m_runtime_cache__)

    def __set__(self, instance, value):
        cache = instance.__m_runtime_cache__
        return self.set(instance, cache, value)

    def __delete__(self, instance):
        if self.name is None:
            raise TypeError("Cannot use cached_property instance without calling __set_name__ on it.")

        if not instance.__m_runtime_cache__:
            msg = f"No '__dict__' attribute on {type(instance).__name__!r} " f"instance to cache {self.name!r} property."
            raise TypeError(msg) from None

        with self._runtime_cache__.__lock_:
            instance.__m_runtime_cache__.pop(self.name, None)


def runtime_cached_property(
    name: str,
    *,
    init_value=None,
    init_cfg_path=None,
    init_at_creation_time=False,
    validity=None,
) -> typing.Callable[[typing.Any], typing.Any]:
    if validity is None:
        validity = cfg_init("default_cache_validity", -1, section="runtime")

    def wrapper(func):
        assert name not in _runtime_cache
        assert name not in _runtime_status

        _runtime_cache[name] = init_value
        _runtime_status[name] = {
            "state": _RuntimeProperty.UNSET,
            "init_at_creation_time": init_at_creation_time,
            "init_cfg_path": init_cfg_path,
            "date": 0,
            "validity": validity,
            "property": _RuntimeProperty(func, name),
        }

        return _runtime_status[name]["property"]

    return wrapper



import errno
import os
import threading
import typing


class LightHandler:
    _handles: dict[int, typing.Any] = dict()
    _key = 4
    _lock = threading.RLock()

    def new(self, *info):
        with self._lock:
            handle = self._key

            self._handles[self._key] = info
            self._key += 1

            return handle

    def _info(self, info, ninfo):
        if info and ninfo is not None:
            try:
                info = info[ninfo]
            except IndexError:
                info = None

        if info is None:
            raise OSError(errno.EBADF, os.strerror(errno.EBADF))

        return info

    def close(self, handle, ninfo=None):
        with self._lock:
            info = self._handles.pop(handle, None)
            return self._info(info, ninfo)

    def h(self, handle, ninfo=None):
        with self._lock:
            info = self._handles.get(handle)
            return self._info(info, ninfo)



from myrrh.core.services.system import ExecutionFailureCauseRVal
from ._mosenv import OsEnv


from myrrh.core.services.system import AbcRuntime
from collections import namedtuple

__mlib__ = "AbcPwd"

struct_passwd = namedtuple("struct_passwd", ("pw_name pw_passwd pw_uid pw_gid pw_gecos pw_dir pw_shell"))


class AbcPwd(AbcRuntime):
    def __init__(self, system):
        super().__init__(system=system)
        self.os = OsEnv(self)

    def getpwuid(self, uid):
        if not isinstance(uid, int):
            raise TypeError("an integer is required")

        pwall = self.getpwall()
        for pw in pwall:
            if pw.pw_uid == uid:
                return pw
        raise KeyError("getpwuid(): uid not found: %s" % uid)

    def getpwnam(self, name):
        if not isinstance(name, str):
            raise TypeError("must be string, not %s" % name.__class__.__name__)

        pwall = self.getpwall()
        for pw in pwall:
            if pw.pw_name == name:
                return pw
        raise KeyError("getpwnam(): name not found: %s" % name)

    def getpwall(self):
        out, err, rval = self.myrrh_os.cmd(b"%(cat)s /etc/passwd")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()

        pwall = []
        for line in out.split(self.os.linesep):
            try:
                line = line.split(":")
                line[2] = int(line[2])
                line[3] = int(line[3])
                pwall.append(self.struct_passwd(*line))
            except (TypeError, ValueError, IndexError):
                pass

        return pwall



import errno
import os
import threading
import weakref

from ..objects import MHandle
from ...interfaces import IRuntimeObject, IProcess


__all__ = ("RuntimeObjectManager",)


class RuntimeObjectManager:
    def __init__(self, max_fds: int | None = None) -> None:
        self.max_fds = max_fds

        self.lock = threading.RLock()
        self.fds: list[IRuntimeObject | None] = list()
        self.procs: dict[int, IProcess] = weakref.WeakValueDictionary()  # type: ignore[assignment]

        if max_fds:
            self.fds.extend([None] * max_fds)

    def get_free_hint(self, count=1):
        with self.lock:
            try:
                first = self.fds.index(None)
                while any(map(lambda o: o is not None, self.fds[first : first + count])):
                    first = self.fds.index(None, first + 1)

                if first + count < len(self.fds):
                    return first

            except ValueError:
                ...

            if not self.max_fds:
                first = len(self.fds)
                self.fds.extend([None] * max(len(self.fds), count * 2))

                return first

            raise OSError(errno.EMFILE, os.strerror(errno.EMFILE))

    def geto(self, hint: int | MHandle) -> IRuntimeObject:
        return weakref.proxy(self._geto(hint))

    def _geto(self, hint) -> IRuntimeObject:
        try:
            obj = self.fds[int(hint)]
            if obj is not None:
                return obj
        except IndexError:
            pass

        raise OSError(errno.EBADF, f"{os.strerror(errno.EBADF)}: {int(hint)}")

    def append(self, *objs: IRuntimeObject) -> int:
        with self.lock:
            hint = self.get_free_hint(count=len(objs))

            for i in range(0, len(objs)):
                obj = objs[i]
                try:
                    self.fds[hint + i] = obj
                except Exception:
                    ...

            return hint

    def register_proc(self, hint):
        with self.lock:
            obj = self._geto(hint)
            self.procs[obj.pid] = obj

    def dup(self, hint: int | MHandle):
        with self.lock:
            hint2 = self.get_free_hint()
            object = self._geto(hint)

            self.fds[hint2] = object

            object.__m_ref_count__ += 1

        return hint2

    def dup2(self, hint: int | MHandle, hint2: int | MHandle):
        with self.lock:
            object = self._geto(hint)

            try:
                self.close(hint2)
            except OSError:
                if OSError.errno == errno.EBADF:
                    pass
                raise

            if not self.max_fds and int(hint2) >= len(self.fds):
                inc = int(hint2) - len(self.fds) + 1
                self.fds.extend([None] * inc)

            self.fds[hint2] = object
            object.__m_ref_count__ += 1

        return hint2

    def close(self, hint: int | MHandle):
        obj = self._geto(hint)
        if not obj.__m_ref_count__ and not obj.closed:
            obj.close()
        else:
            obj.__m_ref_count__ -= 1

        self.fds[int(hint)] = None

    def gethandle(self, hint: int | MHandle) -> MHandle:
        return MHandle(int(hint), self)

    def getrefs(self, hint: int | MHandle) -> tuple[int]:
        with self.lock:
            obj = self._geto(hint)
            return tuple(filter(lambda h: self.fds[h] is obj, range(0, len(self.fds))))  # type: ignore[return-value]

    def getprocs(self, pid=0):
        try:
            return (self.procs[pid],)
        except KeyError:
            pass

        raise ChildProcessError



from .._system.runtime import *  # noqa: F403,F401
from .._system.objects import *  # noqa: F403,F401

from ...provider import Whence, Wiring, Protocol  # noqa: F401

from abc import ABCMeta, ABC

__all__ = (
    "_mlib_",
    "supportedif",
    "ImplPropertyClass",
    "InheritedPropertyClass_properties",
    "InheritedPropertyClass",
)


def _mlib_(impl, default=None) -> type[ABC]:
    try:
        return getattr(impl, impl.__mlib__)
    except AttributeError:
        pass

    if default is not None:
        return default

    raise AttributeError(f"_mlib_ is not defined in {impl}")


# supported decorator
def supportedif(condition, reason=""):
    def condition_wrapper(func):
        def fget(self):
            def wrapper(*a, **kw):
                return func(self, *a, *kw)

            if not condition(self):
                raise AttributeError(reason)
            return wrapper

        return property(fget, doc=func.__doc__)

    return condition_wrapper


class ImplPropertyClass(ABCMeta):
    def __new__(mcls, name, bases, namespace, **kwargs):
        inherited_props = kwargs.get("__inherited_props", tuple())
        inherited_types = kwargs.get("__inherited_types", tuple())

        class Deleted:
            pass

        @property
        def cls_wrapper(self):
            if not hasattr(self, "__propertyclass__%s" % name):
                _bases = bases + tuple(p.fget(self) for p in inherited_props) + tuple(p for p in inherited_types)

                cls = type(name, _bases, namespace)

                gattr = getattr(cls, "__getattr__", None)

                if gattr:

                    def _getattr(_self, name):
                        try:
                            return gattr(_self, name)
                        except AttributeError:
                            pass

                        return getattr(self, name)

                else:

                    def _getattr(_, name):
                        return getattr(self, name)

                setattr(cls, "__getattr__", _getattr)
                cls.__rself__ = self

                setattr(self, "__propertyclass__%s" % name, cls)

            attr = getattr(self, "__propertyclass__%s" % name)

            if attr is Deleted:
                raise AttributeError

            return attr

        @cls_wrapper.setter
        def cls_wrapper(self, value):
            setattr(self, "__propertyclass__%s" % name, value)

        @cls_wrapper.deleter
        def cls_wrapper(self):
            setattr(self, "__propertyclass__%s" % name, Deleted)

        return cls_wrapper


def InheritedPropertyClass_properties(*props):
    class wrapper(ImplPropertyClass):
        def __new__(mcls, name, bases, namespace, **kwargs):
            return super().__new__(mcls, name, bases, namespace, __inherited_props=props)

    return wrapper


def InheritedPropertyClass(*types):
    class wrapper(ImplPropertyClass):
        def __new__(mcls, name, bases, namespace, **kwargs):
            return super().__new__(mcls, name, bases, namespace, __inherited_types=types)

    return wrapper



__all__ = ["System", "Id", "Credentials", "Host", "Supply", "Setting", "GenericItem"]

from ._system import System
from ._id import Id
from ._credentials import Credentials
from ._host import Host
from ._supply import Supply, Settings  # noqa:F401
from .item import GenericItem



# flake8: noqa:

import bmy, unittest

import os
from myrrh.utils.myrrh_test_init import *

setUp(f"bmy.{os.path.basename(__file__)}")


from decimal import Decimal

with bmy.select():
    from mlib.py.test.support import verbose, is_android, is_emscripten, is_wasi

from test.support.warnings_helper import check_warnings
import unittest

with bmy.select():
    from mlib.py import locale
    from mlib.py import sys

import codecs


class BaseLocalizedTest(unittest.TestCase):
    #
    # Base class for tests using a real locale
    #

    @classmethod
    def setUpClass(cls):
        if sys.platform == "darwin":
            with bmy.select():
                from mlib.py import os

            tlocs = ("en_US.UTF-8", "en_US.ISO8859-1", "en_US")
            if int(os.uname().release.split(".")[0]) < 10:
                # The locale test work fine on OSX 10.6, I (ronaldoussoren)
                # haven't had time yet to verify if tests work on OSX 10.5
                # (10.4 is known to be bad)
                raise unittest.SkipTest("Locale support on MacOSX is minimal")
        elif sys.platform.startswith("win"):
            tlocs = ("En", "English")
        else:
            tlocs = ("en_US.UTF-8", "en_US.ISO8859-1", "en_US.US-ASCII", "en_US")
        try:
            oldlocale = locale.setlocale(locale.LC_NUMERIC)
            for tloc in tlocs:
                try:
                    locale.setlocale(locale.LC_NUMERIC, tloc)
                except locale.Error:
                    continue
                break
            else:
                raise unittest.SkipTest("Test locale not supported " "(tried %s)" % (", ".join(tlocs)))
            cls.enUS_locale = tloc
        finally:
            locale.setlocale(locale.LC_NUMERIC, oldlocale)

    def setUp(self):
        oldlocale = locale.setlocale(self.locale_type)
        self.addCleanup(locale.setlocale, self.locale_type, oldlocale)
        locale.setlocale(self.locale_type, self.enUS_locale)
        if verbose:
            print("testing with %r..." % self.enUS_locale, end=" ", flush=True)


class BaseCookedTest(unittest.TestCase):
    #
    # Base class for tests using cooked localeconv() values
    #

    def setUp(self):
        locale._override_localeconv = self.cooked_values

    def tearDown(self):
        locale._override_localeconv = {}


class CCookedTest(BaseCookedTest):
    # A cooked "C" locale

    cooked_values = {
        "currency_symbol": "",
        "decimal_point": ".",
        "frac_digits": 127,
        "grouping": [],
        "int_curr_symbol": "",
        "int_frac_digits": 127,
        "mon_decimal_point": "",
        "mon_grouping": [],
        "mon_thousands_sep": "",
        "n_cs_precedes": 127,
        "n_sep_by_space": 127,
        "n_sign_posn": 127,
        "negative_sign": "",
        "p_cs_precedes": 127,
        "p_sep_by_space": 127,
        "p_sign_posn": 127,
        "positive_sign": "",
        "thousands_sep": "",
    }


class EnUSCookedTest(BaseCookedTest):
    # A cooked "en_US" locale

    cooked_values = {
        "currency_symbol": "$",
        "decimal_point": ".",
        "frac_digits": 2,
        "grouping": [3, 3, 0],
        "int_curr_symbol": "USD ",
        "int_frac_digits": 2,
        "mon_decimal_point": ".",
        "mon_grouping": [3, 3, 0],
        "mon_thousands_sep": ",",
        "n_cs_precedes": 1,
        "n_sep_by_space": 0,
        "n_sign_posn": 1,
        "negative_sign": "-",
        "p_cs_precedes": 1,
        "p_sep_by_space": 0,
        "p_sign_posn": 1,
        "positive_sign": "",
        "thousands_sep": ",",
    }


class FrFRCookedTest(BaseCookedTest):
    # A cooked "fr_FR" locale with a space character as decimal separator
    # and a non-ASCII currency symbol.

    cooked_values = {
        "currency_symbol": "\u20ac",
        "decimal_point": ",",
        "frac_digits": 2,
        "grouping": [3, 3, 0],
        "int_curr_symbol": "EUR ",
        "int_frac_digits": 2,
        "mon_decimal_point": ",",
        "mon_grouping": [3, 3, 0],
        "mon_thousands_sep": " ",
        "n_cs_precedes": 0,
        "n_sep_by_space": 1,
        "n_sign_posn": 1,
        "negative_sign": "-",
        "p_cs_precedes": 0,
        "p_sep_by_space": 1,
        "p_sign_posn": 1,
        "positive_sign": "",
        "thousands_sep": " ",
    }


class BaseFormattingTest(object):
    #
    # Utility functions for formatting tests
    #

    def _test_formatfunc(self, format, value, out, func, **format_opts):
        self.assertEqual(func(format, value, **format_opts), out)

    def _test_format(self, format, value, out, **format_opts):
        with check_warnings(("", DeprecationWarning)):
            self._test_formatfunc(format, value, out, func=locale.format, **format_opts)

    def _test_format_string(self, format, value, out, **format_opts):
        self._test_formatfunc(format, value, out, func=locale.format_string, **format_opts)

    def _test_currency(self, value, out, **format_opts):
        self.assertEqual(locale.currency(value, **format_opts), out)


class EnUSNumberFormatting(BaseFormattingTest):
    # XXX there is a grouping + padding bug when the thousands separator
    # is empty but the grouping array contains values (e.g. Solaris 10)

    def setUp(self):
        self.sep = locale.localeconv()["thousands_sep"]

    def test_grouping(self):
        self._test_format("%f", 1024, grouping=1, out="1%s024.000000" % self.sep)
        self._test_format("%f", 102, grouping=1, out="102.000000")
        self._test_format("%f", -42, grouping=1, out="-42.000000")
        self._test_format("%+f", -42, grouping=1, out="-42.000000")

    def test_grouping_and_padding(self):
        self._test_format("%20.f", -42, grouping=1, out="-42".rjust(20))
        if self.sep:
            self._test_format("%+10.f", -4200, grouping=1, out=("-4%s200" % self.sep).rjust(10))
            self._test_format("%-10.f", -4200, grouping=1, out=("-4%s200" % self.sep).ljust(10))

    def test_integer_grouping(self):
        self._test_format("%d", 4200, grouping=True, out="4%s200" % self.sep)
        self._test_format("%+d", 4200, grouping=True, out="+4%s200" % self.sep)
        self._test_format("%+d", -4200, grouping=True, out="-4%s200" % self.sep)

    def test_integer_grouping_and_padding(self):
        self._test_format("%10d", 4200, grouping=True, out=("4%s200" % self.sep).rjust(10))
        self._test_format("%-10d", -4200, grouping=True, out=("-4%s200" % self.sep).ljust(10))

    def test_simple(self):
        self._test_format("%f", 1024, grouping=0, out="1024.000000")
        self._test_format("%f", 102, grouping=0, out="102.000000")
        self._test_format("%f", -42, grouping=0, out="-42.000000")
        self._test_format("%+f", -42, grouping=0, out="-42.000000")

    def test_padding(self):
        self._test_format("%20.f", -42, grouping=0, out="-42".rjust(20))
        self._test_format("%+10.f", -4200, grouping=0, out="-4200".rjust(10))
        self._test_format("%-10.f", 4200, grouping=0, out="4200".ljust(10))

    def test_format_deprecation(self):
        with self.assertWarns(DeprecationWarning):
            locale.format("%-10.f", 4200, grouping=True)

    def test_complex_formatting(self):
        # Spaces in formatting string
        self._test_format_string(
            "One million is %i",
            1000000,
            grouping=1,
            out="One million is 1%s000%s000" % (self.sep, self.sep),
        )
        self._test_format_string(
            "One  million is %i",
            1000000,
            grouping=1,
            out="One  million is 1%s000%s000" % (self.sep, self.sep),
        )
        # Dots in formatting string
        self._test_format_string(".%f.", 1000.0, out=".1000.000000.")
        # Padding
        if self.sep:
            self._test_format_string(
                "-->  %10.2f",
                4200,
                grouping=1,
                out="-->  " + ("4%s200.00" % self.sep).rjust(10),
            )
        # Asterisk formats
        self._test_format_string("%10.*f", (2, 1000), grouping=0, out="1000.00".rjust(10))
        if self.sep:
            self._test_format_string(
                "%*.*f",
                (10, 2, 1000),
                grouping=1,
                out=("1%s000.00" % self.sep).rjust(10),
            )
        # Test more-in-one
        if self.sep:
            self._test_format_string(
                "int %i float %.2f str %s",
                (1000, 1000.0, "str"),
                grouping=1,
                out="int 1%s000 float 1%s000.00 str str" % (self.sep, self.sep),
            )


class TestFormatPatternArg(unittest.TestCase):
    # Test handling of pattern argument of format

    def test_onlyOnePattern(self):
        with check_warnings(("", DeprecationWarning)):
            # Issue 2522: accept exactly one % pattern, and no extra chars.
            self.assertRaises(ValueError, locale.format, "%f\n", "foo")
            self.assertRaises(ValueError, locale.format, "%f\r", "foo")
            self.assertRaises(ValueError, locale.format, "%f\r\n", "foo")
            self.assertRaises(ValueError, locale.format, " %f", "foo")
            self.assertRaises(ValueError, locale.format, "%fg", "foo")
            self.assertRaises(ValueError, locale.format, "%^g", "foo")
            self.assertRaises(ValueError, locale.format, "%f%%", "foo")


class TestLocaleFormatString(unittest.TestCase):
    """General tests on locale.format_string"""

    def test_percent_escape(self):
        self.assertEqual(locale.format_string("%f%%", 1.0), "%f%%" % 1.0)
        self.assertEqual(locale.format_string("%d %f%%d", (1, 1.0)), "%d %f%%d" % (1, 1.0))
        self.assertEqual(
            locale.format_string("%(foo)s %%d", {"foo": "bar"}),
            ("%(foo)s %%d" % {"foo": "bar"}),
        )

    def test_mapping(self):
        self.assertEqual(
            locale.format_string("%(foo)s bing.", {"foo": "bar"}),
            ("%(foo)s bing." % {"foo": "bar"}),
        )
        self.assertEqual(
            locale.format_string("%(foo)s", {"foo": "bar"}),
            ("%(foo)s" % {"foo": "bar"}),
        )


class TestNumberFormatting(BaseLocalizedTest, EnUSNumberFormatting):
    # Test number formatting with a real English locale.

    locale_type = locale.LC_NUMERIC

    def setUp(self):
        BaseLocalizedTest.setUp(self)
        EnUSNumberFormatting.setUp(self)


class TestEnUSNumberFormatting(EnUSCookedTest, EnUSNumberFormatting):
    # Test number formatting with a cooked "en_US" locale.

    def setUp(self):
        EnUSCookedTest.setUp(self)
        EnUSNumberFormatting.setUp(self)

    def test_currency(self):
        self._test_currency(50000, "$50000.00")
        self._test_currency(50000, "$50,000.00", grouping=True)
        self._test_currency(50000, "USD 50,000.00", grouping=True, international=True)


class TestCNumberFormatting(CCookedTest, BaseFormattingTest):
    # Test number formatting with a cooked "C" locale.

    def test_grouping(self):
        self._test_format("%.2f", 12345.67, grouping=True, out="12345.67")

    def test_grouping_and_padding(self):
        self._test_format("%9.2f", 12345.67, grouping=True, out=" 12345.67")


class TestFrFRNumberFormatting(FrFRCookedTest, BaseFormattingTest):
    # Test number formatting with a cooked "fr_FR" locale.

    def test_decimal_point(self):
        self._test_format("%.2f", 12345.67, out="12345,67")

    def test_grouping(self):
        self._test_format("%.2f", 345.67, grouping=True, out="345,67")
        self._test_format("%.2f", 12345.67, grouping=True, out="12 345,67")

    def test_grouping_and_padding(self):
        self._test_format("%6.2f", 345.67, grouping=True, out="345,67")
        self._test_format("%7.2f", 345.67, grouping=True, out=" 345,67")
        self._test_format("%8.2f", 12345.67, grouping=True, out="12 345,67")
        self._test_format("%9.2f", 12345.67, grouping=True, out="12 345,67")
        self._test_format("%10.2f", 12345.67, grouping=True, out=" 12 345,67")
        self._test_format("%-6.2f", 345.67, grouping=True, out="345,67")
        self._test_format("%-7.2f", 345.67, grouping=True, out="345,67 ")
        self._test_format("%-8.2f", 12345.67, grouping=True, out="12 345,67")
        self._test_format("%-9.2f", 12345.67, grouping=True, out="12 345,67")
        self._test_format("%-10.2f", 12345.67, grouping=True, out="12 345,67 ")

    def test_integer_grouping(self):
        self._test_format("%d", 200, grouping=True, out="200")
        self._test_format("%d", 4200, grouping=True, out="4 200")

    def test_integer_grouping_and_padding(self):
        self._test_format("%4d", 4200, grouping=True, out="4 200")
        self._test_format("%5d", 4200, grouping=True, out="4 200")
        self._test_format("%10d", 4200, grouping=True, out="4 200".rjust(10))
        self._test_format("%-4d", 4200, grouping=True, out="4 200")
        self._test_format("%-5d", 4200, grouping=True, out="4 200")
        self._test_format("%-10d", 4200, grouping=True, out="4 200".ljust(10))

    def test_currency(self):
        euro = "\u20ac"
        self._test_currency(50000, "50000,00 " + euro)
        self._test_currency(50000, "50 000,00 " + euro, grouping=True)
        self._test_currency(50000, "50 000,00 EUR", grouping=True, international=True)


class TestCollation(unittest.TestCase):
    # Test string collation functions

    def test_strcoll(self):
        self.assertLess(locale.strcoll("a", "b"), 0)
        self.assertEqual(locale.strcoll("a", "a"), 0)
        self.assertGreater(locale.strcoll("b", "a"), 0)
        # embedded null character
        self.assertRaises(ValueError, locale.strcoll, "a\0", "a")
        self.assertRaises(ValueError, locale.strcoll, "a", "a\0")

    def test_strxfrm(self):
        self.assertLess(locale.strxfrm("a"), locale.strxfrm("b"))
        # embedded null character
        self.assertRaises(ValueError, locale.strxfrm, "a\0")


class TestEnUSCollation(BaseLocalizedTest, TestCollation):
    # Test string collation functions with a real English locale

    locale_type = locale.LC_ALL

    def setUp(self):
        enc = codecs.lookup(locale.getencoding() or "ascii").name
        if enc not in ("utf-8", "iso8859-1", "cp1252"):
            raise unittest.SkipTest("encoding not suitable")
        if enc != "iso8859-1" and (sys.platform == "darwin" or is_android or sys.platform.startswith("freebsd")):
            raise unittest.SkipTest("wcscoll/wcsxfrm have known bugs")
        BaseLocalizedTest.setUp(self)

    @unittest.skipIf(sys.platform.startswith("aix"), "bpo-29972: broken test on AIX")
    @unittest.skipIf(is_emscripten or is_wasi, "musl libc issue on Emscripten/WASI, bpo-46390")
    def test_strcoll_with_diacritic(self):
        self.assertLess(locale.strcoll("à", "b"), 0)

    @unittest.skipIf(sys.platform.startswith("aix"), "bpo-29972: broken test on AIX")
    @unittest.skipIf(is_emscripten or is_wasi, "musl libc issue on Emscripten/WASI, bpo-46390")
    def test_strxfrm_with_diacritic(self):
        self.assertLess(locale.strxfrm("à"), locale.strxfrm("b"))


class NormalizeTest(unittest.TestCase):
    def check(self, localename, expected):
        self.assertEqual(locale.normalize(localename), expected, msg=localename)

    def test_locale_alias(self):
        for localename, alias in locale.locale_alias.items():
            with self.subTest(locale=(localename, alias)):
                self.check(localename, alias)

    def test_empty(self):
        self.check("", "")

    def test_c(self):
        self.check("c", "C")
        self.check("posix", "C")

    def test_english(self):
        self.check("en", "en_US.ISO8859-1")
        self.check("EN", "en_US.ISO8859-1")
        self.check("en.iso88591", "en_US.ISO8859-1")
        self.check("en_US", "en_US.ISO8859-1")
        self.check("en_us", "en_US.ISO8859-1")
        self.check("en_GB", "en_GB.ISO8859-1")
        self.check("en_US.UTF-8", "en_US.UTF-8")
        self.check("en_US.utf8", "en_US.UTF-8")
        self.check("en_US:UTF-8", "en_US.UTF-8")
        self.check("en_US.ISO8859-1", "en_US.ISO8859-1")
        self.check("en_US.US-ASCII", "en_US.ISO8859-1")
        self.check("en_US.88591", "en_US.ISO8859-1")
        self.check("en_US.885915", "en_US.ISO8859-15")
        self.check("english", "en_EN.ISO8859-1")
        self.check("english_uk.ascii", "en_GB.ISO8859-1")

    def test_hyphenated_encoding(self):
        self.check("az_AZ.iso88599e", "az_AZ.ISO8859-9E")
        self.check("az_AZ.ISO8859-9E", "az_AZ.ISO8859-9E")
        self.check("tt_RU.koi8c", "tt_RU.KOI8-C")
        self.check("tt_RU.KOI8-C", "tt_RU.KOI8-C")
        self.check("lo_LA.cp1133", "lo_LA.IBM-CP1133")
        self.check("lo_LA.ibmcp1133", "lo_LA.IBM-CP1133")
        self.check("lo_LA.IBM-CP1133", "lo_LA.IBM-CP1133")
        self.check("uk_ua.microsoftcp1251", "uk_UA.CP1251")
        self.check("uk_ua.microsoft-cp1251", "uk_UA.CP1251")
        self.check("ka_ge.georgianacademy", "ka_GE.GEORGIAN-ACADEMY")
        self.check("ka_GE.GEORGIAN-ACADEMY", "ka_GE.GEORGIAN-ACADEMY")
        self.check("cs_CZ.iso88592", "cs_CZ.ISO8859-2")
        self.check("cs_CZ.ISO8859-2", "cs_CZ.ISO8859-2")

    def test_euro_modifier(self):
        self.check("de_DE@euro", "de_DE.ISO8859-15")
        self.check("en_US.ISO8859-15@euro", "en_US.ISO8859-15")
        self.check("de_DE.utf8@euro", "de_DE.UTF-8")

    def test_latin_modifier(self):
        self.check("be_BY.UTF-8@latin", "be_BY.UTF-8@latin")
        self.check("sr_RS.UTF-8@latin", "sr_RS.UTF-8@latin")
        self.check("sr_RS.UTF-8@latn", "sr_RS.UTF-8@latin")

    def test_valencia_modifier(self):
        self.check("ca_ES.UTF-8@valencia", "ca_ES.UTF-8@valencia")
        self.check("ca_ES@valencia", "ca_ES.UTF-8@valencia")
        self.check("ca@valencia", "ca_ES.ISO8859-1@valencia")

    def test_devanagari_modifier(self):
        self.check("ks_IN.UTF-8@devanagari", "ks_IN.UTF-8@devanagari")
        self.check("ks_IN@devanagari", "ks_IN.UTF-8@devanagari")
        self.check("ks@devanagari", "ks_IN.UTF-8@devanagari")
        self.check("ks_IN.UTF-8", "ks_IN.UTF-8")
        self.check("ks_IN", "ks_IN.UTF-8")
        self.check("ks", "ks_IN.UTF-8")
        self.check("sd_IN.UTF-8@devanagari", "sd_IN.UTF-8@devanagari")
        self.check("sd_IN@devanagari", "sd_IN.UTF-8@devanagari")
        self.check("sd@devanagari", "sd_IN.UTF-8@devanagari")
        self.check("sd_IN.UTF-8", "sd_IN.UTF-8")
        self.check("sd_IN", "sd_IN.UTF-8")
        self.check("sd", "sd_IN.UTF-8")

    def test_euc_encoding(self):
        self.check("ja_jp.euc", "ja_JP.eucJP")
        self.check("ja_jp.eucjp", "ja_JP.eucJP")
        self.check("ko_kr.euc", "ko_KR.eucKR")
        self.check("ko_kr.euckr", "ko_KR.eucKR")
        self.check("zh_cn.euc", "zh_CN.eucCN")
        self.check("zh_tw.euc", "zh_TW.eucTW")
        self.check("zh_tw.euctw", "zh_TW.eucTW")

    def test_japanese(self):
        self.check("ja", "ja_JP.eucJP")
        self.check("ja.jis", "ja_JP.JIS7")
        self.check("ja.sjis", "ja_JP.SJIS")
        self.check("ja_jp", "ja_JP.eucJP")
        self.check("ja_jp.ajec", "ja_JP.eucJP")
        self.check("ja_jp.euc", "ja_JP.eucJP")
        self.check("ja_jp.eucjp", "ja_JP.eucJP")
        self.check("ja_jp.iso-2022-jp", "ja_JP.JIS7")
        self.check("ja_jp.iso2022jp", "ja_JP.JIS7")
        self.check("ja_jp.jis", "ja_JP.JIS7")
        self.check("ja_jp.jis7", "ja_JP.JIS7")
        self.check("ja_jp.mscode", "ja_JP.SJIS")
        self.check("ja_jp.pck", "ja_JP.SJIS")
        self.check("ja_jp.sjis", "ja_JP.SJIS")
        self.check("ja_jp.ujis", "ja_JP.eucJP")
        self.check("ja_jp.utf8", "ja_JP.UTF-8")
        self.check("japan", "ja_JP.eucJP")
        self.check("japanese", "ja_JP.eucJP")
        self.check("japanese-euc", "ja_JP.eucJP")
        self.check("japanese.euc", "ja_JP.eucJP")
        self.check("japanese.sjis", "ja_JP.SJIS")
        self.check("jp_jp", "ja_JP.eucJP")


class TestMiscellaneous(unittest.TestCase):
    def test_defaults_UTF8(self):
        # Issue #18378: on (at least) macOS setting LC_CTYPE to "UTF-8" is
        # valid. Furthermore LC_CTYPE=UTF is used by the UTF-8 locale coercing
        # during interpreter startup (on macOS).
        with bmy.select():
            _locale = None
            from mlib.py import os

        self.assertEqual(locale._parse_localename("UTF-8"), (None, "UTF-8"))

        if hasattr(_locale, "_getdefaultlocale"):
            orig_getlocale = _locale._getdefaultlocale
            del _locale._getdefaultlocale
        else:
            orig_getlocale = None

        orig_env = {}
        try:
            for key in ("LC_ALL", "LC_CTYPE", "LANG", "LANGUAGE"):
                if key in os.environ:
                    orig_env[key] = os.environ[key]
                    del os.environ[key]

            os.environ["LC_CTYPE"] = "UTF-8"

            with check_warnings(("", DeprecationWarning)):
                self.assertEqual(locale.getdefaultlocale(), (None, "UTF-8"))

        finally:
            for k in orig_env:
                os.environ[k] = orig_env[k]

            if "LC_CTYPE" not in orig_env:
                del os.environ["LC_CTYPE"]

            if orig_getlocale is not None:
                _locale._getdefaultlocale = orig_getlocale

    def test_getencoding(self):
        # Invoke getencoding to make sure it does not cause exceptions.
        enc = locale.getencoding()
        self.assertIsInstance(enc, str)
        self.assertNotEqual(enc, "")
        # make sure it is valid
        codecs.lookup(enc)

    def test_getpreferredencoding(self):
        # Invoke getpreferredencoding to make sure it does not cause exceptions.
        enc = locale.getpreferredencoding()
        if enc:
            # If encoding non-empty, make sure it is valid
            codecs.lookup(enc)

    def test_strcoll_3303(self):
        # test crasher from bug #3303
        self.assertRaises(TypeError, locale.strcoll, "a", None)
        self.assertRaises(TypeError, locale.strcoll, b"a", None)

    def test_setlocale_category(self):
        locale.setlocale(locale.LC_ALL)
        locale.setlocale(locale.LC_TIME)
        locale.setlocale(locale.LC_CTYPE)
        locale.setlocale(locale.LC_COLLATE)
        locale.setlocale(locale.LC_MONETARY)
        locale.setlocale(locale.LC_NUMERIC)

        # crasher from bug #7419
        self.assertRaises(locale.Error, locale.setlocale, 12345)

    def test_getsetlocale_issue1813(self):
        # Issue #1813: setting and getting the locale under a Turkish locale
        oldlocale = locale.setlocale(locale.LC_CTYPE)
        self.addCleanup(locale.setlocale, locale.LC_CTYPE, oldlocale)
        try:
            locale.setlocale(locale.LC_CTYPE, "tr_TR")
        except locale.Error:
            # Unsupported locale on this system
            self.skipTest("test needs Turkish locale")
        loc = locale.getlocale(locale.LC_CTYPE)
        if verbose:
            print("testing with %a" % (loc,), end=" ", flush=True)
        try:
            locale.setlocale(locale.LC_CTYPE, loc)
        except locale.Error as exc:
            # bpo-37945: setlocale(LC_CTYPE) fails with getlocale(LC_CTYPE)
            # and the tr_TR locale on Windows. getlocale() builds a locale
            # which is not recognize by setlocale().
            self.skipTest(f"setlocale(LC_CTYPE, {loc!r}) failed: {exc!r}")
        self.assertEqual(loc, locale.getlocale(locale.LC_CTYPE))

    def test_invalid_locale_format_in_localetuple(self):
        with self.assertRaises(TypeError):
            locale.setlocale(locale.LC_ALL, b"fi_FI")

    def test_invalid_iterable_in_localetuple(self):
        with self.assertRaises(TypeError):
            locale.setlocale(locale.LC_ALL, (b"not", b"valid"))


class BaseDelocalizeTest(BaseLocalizedTest):
    def _test_delocalize(self, value, out):
        self.assertEqual(locale.delocalize(value), out)

    def _test_atof(self, value, out):
        self.assertEqual(locale.atof(value), out)

    def _test_atoi(self, value, out):
        self.assertEqual(locale.atoi(value), out)


class TestEnUSDelocalize(EnUSCookedTest, BaseDelocalizeTest):
    def test_delocalize(self):
        self._test_delocalize("50000.00", "50000.00")
        self._test_delocalize("50,000.00", "50000.00")

    def test_atof(self):
        self._test_atof("50000.00", 50000.0)
        self._test_atof("50,000.00", 50000.0)

    def test_atoi(self):
        self._test_atoi("50000", 50000)
        self._test_atoi("50,000", 50000)


class TestCDelocalizeTest(CCookedTest, BaseDelocalizeTest):
    def test_delocalize(self):
        self._test_delocalize("50000.00", "50000.00")

    def test_atof(self):
        self._test_atof("50000.00", 50000.0)

    def test_atoi(self):
        self._test_atoi("50000", 50000)


class TestfrFRDelocalizeTest(FrFRCookedTest, BaseDelocalizeTest):
    def test_delocalize(self):
        self._test_delocalize("50000,00", "50000.00")
        self._test_delocalize("50 000,00", "50000.00")

    def test_atof(self):
        self._test_atof("50000,00", 50000.0)
        self._test_atof("50 000,00", 50000.0)

    def test_atoi(self):
        self._test_atoi("50000", 50000)
        self._test_atoi("50 000", 50000)


class BaseLocalizeTest(BaseLocalizedTest):
    def _test_localize(self, value, out, grouping=False):
        self.assertEqual(locale._localize(value, grouping=grouping), out)


class TestEnUSLocalize(EnUSCookedTest, BaseLocalizeTest):
    def test_localize(self):
        self._test_localize("50000.00", "50000.00")
        self._test_localize("{0:.16f}".format(Decimal("1.15")), "1.1500000000000000")


class TestCLocalize(CCookedTest, BaseLocalizeTest):
    def test_localize(self):
        self._test_localize("50000.00", "50000.00")


class TestfrFRLocalize(FrFRCookedTest, BaseLocalizeTest):
    def test_localize(self):
        self._test_localize("50000.00", "50000,00")
        self._test_localize("50000.00", "50 000,00", grouping=True)


if __name__ == "__main__":
    unittest.main()



# flake8: noqa:
"""Test cases for the fnmatch module."""


import bmy, unittest

import os
from myrrh.utils.myrrh_test_init import *

setUp(f"bmy.{os.path.basename(__file__)}")

import string
import warnings

with bmy.select():
    from mlib.py import os
    from mlib.py.fnmatch import fnmatch, fnmatchcase, translate, filter


class FnmatchTestCase(unittest.TestCase):
    def check_match(self, filename, pattern, should_match=True, fn=fnmatch):
        if should_match:
            self.assertTrue(
                fn(filename, pattern),
                "expected %r to match pattern %r" % (filename, pattern),
            )
        else:
            self.assertFalse(
                fn(filename, pattern),
                "expected %r not to match pattern %r" % (filename, pattern),
            )

    def test_fnmatch(self):
        check = self.check_match
        check("abc", "abc")
        check("abc", "?*?")
        check("abc", "???*")
        check("abc", "*???")
        check("abc", "???")
        check("abc", "*")
        check("abc", "ab[cd]")
        check("abc", "ab[!de]")
        check("abc", "ab[de]", False)
        check("a", "??", False)
        check("a", "b", False)

        # these test that '\' is handled correctly in character sets;
        # see SF bug #409651
        check("\\", r"[\]")
        check("a", r"[!\]")
        check("\\", r"[!\]", False)

        # test that filenames with newlines in them are handled correctly.
        # http://bugs.python.org/issue6665
        check("foo\nbar", "foo*")
        check("foo\nbar\n", "foo*")
        check("\nfoo", "foo*", False)
        check("\n", "*")

    def test_slow_fnmatch(self):
        check = self.check_match
        check("a" * 50, "*a*a*a*a*a*a*a*a*a*a")
        # The next "takes forever" if the regexp translation is
        # straightforward.  See bpo-40480.
        check("a" * 50 + "b", "*a*a*a*a*a*a*a*a*a*a", False)

    def test_mix_bytes_str(self):
        self.assertRaises(TypeError, fnmatch, "test", b"*")
        self.assertRaises(TypeError, fnmatch, b"test", "*")
        self.assertRaises(TypeError, fnmatchcase, "test", b"*")
        self.assertRaises(TypeError, fnmatchcase, b"test", "*")

    def test_fnmatchcase(self):
        check = self.check_match
        check("abc", "abc", True, fnmatchcase)
        check("AbC", "abc", False, fnmatchcase)
        check("abc", "AbC", False, fnmatchcase)
        check("AbC", "AbC", True, fnmatchcase)

        check("usr/bin", "usr/bin", True, fnmatchcase)
        check("usr\\bin", "usr/bin", False, fnmatchcase)
        check("usr/bin", "usr\\bin", False, fnmatchcase)
        check("usr\\bin", "usr\\bin", True, fnmatchcase)

    def test_bytes(self):
        self.check_match(b"test", b"te*")
        self.check_match(b"test\xff", b"te*\xff")
        self.check_match(b"foo\nbar", b"foo*")

    def test_case(self):
        ignorecase = os.path.normcase("ABC") == os.path.normcase("abc")
        check = self.check_match
        check("abc", "abc")
        check("AbC", "abc", ignorecase)
        check("abc", "AbC", ignorecase)
        check("AbC", "AbC")

    def test_sep(self):
        normsep = os.path.normcase("\\") == os.path.normcase("/")
        check = self.check_match
        check("usr/bin", "usr/bin")
        check("usr\\bin", "usr/bin", normsep)
        check("usr/bin", "usr\\bin", normsep)
        check("usr\\bin", "usr\\bin")

    def test_char_set(self):
        ignorecase = os.path.normcase("ABC") == os.path.normcase("abc")
        check = self.check_match
        testcases = string.ascii_lowercase + string.digits + string.punctuation
        for c in testcases:
            check(c, "[az]", c in "az")
            check(c, "[!az]", c not in "az")
        # Case insensitive.
        for c in testcases:
            check(c, "[AZ]", (c in "az") and ignorecase)
            check(c, "[!AZ]", (c not in "az") or not ignorecase)
        for c in string.ascii_uppercase:
            check(c, "[az]", (c in "AZ") and ignorecase)
            check(c, "[!az]", (c not in "AZ") or not ignorecase)
        # Repeated same character.
        for c in testcases:
            check(c, "[aa]", c == "a")
        # Special cases.
        for c in testcases:
            check(c, "[^az]", c in "^az")
            check(c, "[[az]", c in "[az")
            check(c, r"[!]]", c != "]")
        check("[", "[")
        check("[]", "[]")
        check("[!", "[!")
        check("[!]", "[!]")

    def test_range(self):
        ignorecase = os.path.normcase("ABC") == os.path.normcase("abc")
        normsep = os.path.normcase("\\") == os.path.normcase("/")
        check = self.check_match
        testcases = string.ascii_lowercase + string.digits + string.punctuation
        for c in testcases:
            check(c, "[b-d]", c in "bcd")
            check(c, "[!b-d]", c not in "bcd")
            check(c, "[b-dx-z]", c in "bcdxyz")
            check(c, "[!b-dx-z]", c not in "bcdxyz")
        # Case insensitive.
        for c in testcases:
            check(c, "[B-D]", (c in "bcd") and ignorecase)
            check(c, "[!B-D]", (c not in "bcd") or not ignorecase)
        for c in string.ascii_uppercase:
            check(c, "[b-d]", (c in "BCD") and ignorecase)
            check(c, "[!b-d]", (c not in "BCD") or not ignorecase)
        # Upper bound == lower bound.
        for c in testcases:
            check(c, "[b-b]", c == "b")
        # Special cases.
        for c in testcases:
            check(c, "[!-#]", c not in "-#")
            check(c, "[!--.]", c not in "-.")
            check(c, "[^-`]", c in "^_`")
            if not (normsep and c == "/"):
                check(c, "[[-^]", c in r"[\]^")
                check(c, r"[\-^]", c in r"\]^")
            check(c, "[b-]", c in "-b")
            check(c, "[!b-]", c not in "-b")
            check(c, "[-b]", c in "-b")
            check(c, "[!-b]", c not in "-b")
            check(c, "[-]", c in "-")
            check(c, "[!-]", c not in "-")
        # Upper bound is less that lower bound: error in RE.
        for c in testcases:
            check(c, "[d-b]", False)
            check(c, "[!d-b]", True)
            check(c, "[d-bx-z]", c in "xyz")
            check(c, "[!d-bx-z]", c not in "xyz")
            check(c, "[d-b^-`]", c in "^_`")
            if not (normsep and c == "/"):
                check(c, "[d-b[-^]", c in r"[\]^")

    def test_sep_in_char_set(self):
        normsep = os.path.normcase("\\") == os.path.normcase("/")
        check = self.check_match
        check("/", r"[/]")
        check("\\", r"[\]")
        check("/", r"[\]", normsep)
        check("\\", r"[/]", normsep)
        check("[/]", r"[/]", False)
        check(r"[\\]", r"[/]", False)
        check("\\", r"[\t]")
        check("/", r"[\t]", normsep)
        check("t", r"[\t]")
        check("\t", r"[\t]", False)

    def test_sep_in_range(self):
        normsep = os.path.normcase("\\") == os.path.normcase("/")
        check = self.check_match
        check("a/b", "a[.-0]b", not normsep)
        check("a\\b", "a[.-0]b", False)
        check("a\\b", "a[Z-^]b", not normsep)
        check("a/b", "a[Z-^]b", False)

        check("a/b", "a[/-0]b", not normsep)
        check(r"a\b", "a[/-0]b", False)
        check("a[/-0]b", "a[/-0]b", False)
        check(r"a[\-0]b", "a[/-0]b", False)

        check("a/b", "a[.-/]b")
        check(r"a\b", "a[.-/]b", normsep)
        check("a[.-/]b", "a[.-/]b", False)
        check(r"a[.-\]b", "a[.-/]b", False)

        check(r"a\b", r"a[\-^]b")
        check("a/b", r"a[\-^]b", normsep)
        check(r"a[\-^]b", r"a[\-^]b", False)
        check("a[/-^]b", r"a[\-^]b", False)

        check(r"a\b", r"a[Z-\]b", not normsep)
        check("a/b", r"a[Z-\]b", False)
        check(r"a[Z-\]b", r"a[Z-\]b", False)
        check("a[Z-/]b", r"a[Z-\]b", False)

    def test_warnings(self):
        with warnings.catch_warnings():
            warnings.simplefilter("error", Warning)
            check = self.check_match
            check("[", "[[]")
            check("&", "[a&&b]")
            check("|", "[a||b]")
            check("~", "[a~~b]")
            check(",", "[a-z+--A-Z]")
            check(".", "[a-z--/A-Z]")


class TranslateTestCase(unittest.TestCase):
    def test_translate(self):
        import re

        self.assertEqual(translate("*"), r"(?s:.*)\Z")
        self.assertEqual(translate("?"), r"(?s:.)\Z")
        self.assertEqual(translate("a?b*"), r"(?s:a.b.*)\Z")
        self.assertEqual(translate("[abc]"), r"(?s:[abc])\Z")
        self.assertEqual(translate("[]]"), r"(?s:[]])\Z")
        self.assertEqual(translate("[!x]"), r"(?s:[^x])\Z")
        self.assertEqual(translate("[^x]"), r"(?s:[\^x])\Z")
        self.assertEqual(translate("[x"), r"(?s:\[x)\Z")
        # from the docs
        self.assertEqual(translate("*.txt"), r"(?s:.*\.txt)\Z")
        # squash consecutive stars
        self.assertEqual(translate("*********"), r"(?s:.*)\Z")
        self.assertEqual(translate("A*********"), r"(?s:A.*)\Z")
        self.assertEqual(translate("*********A"), r"(?s:.*A)\Z")
        self.assertEqual(translate("A*********?[?]?"), r"(?s:A.*.[?].)\Z")
        # fancy translation to prevent exponential-time match failure
        t = translate("**a*a****a")
        self.assertEqual(t, r"(?s:(?>.*?a)(?>.*?a).*a)\Z")
        # and try pasting multiple translate results - it's an undocumented
        # feature that this works
        r1 = translate("**a**a**a*")
        r2 = translate("**b**b**b*")
        r3 = translate("*c*c*c*")
        fatre = "|".join([r1, r2, r3])
        self.assertTrue(re.match(fatre, "abaccad"))
        self.assertTrue(re.match(fatre, "abxbcab"))
        self.assertTrue(re.match(fatre, "cbabcaxc"))
        self.assertFalse(re.match(fatre, "dabccbad"))


class FilterTestCase(unittest.TestCase):
    def test_filter(self):
        self.assertEqual(filter(["Python", "Ruby", "Perl", "Tcl"], "P*"), ["Python", "Perl"])
        self.assertEqual(filter([b"Python", b"Ruby", b"Perl", b"Tcl"], b"P*"), [b"Python", b"Perl"])

    def test_mix_bytes_str(self):
        self.assertRaises(TypeError, filter, ["test"], b"*")
        self.assertRaises(TypeError, filter, [b"test"], "*")

    def test_case(self):
        ignorecase = os.path.normcase("P") == os.path.normcase("p")
        self.assertEqual(
            filter(["Test.py", "Test.rb", "Test.PL"], "*.p*"),
            ["Test.py", "Test.PL"] if ignorecase else ["Test.py"],
        )
        self.assertEqual(
            filter(["Test.py", "Test.rb", "Test.PL"], "*.P*"),
            ["Test.py", "Test.PL"] if ignorecase else ["Test.PL"],
        )

    def test_sep(self):
        normsep = os.path.normcase("\\") == os.path.normcase("/")
        self.assertEqual(
            filter(["usr/bin", "usr", "usr\\lib"], "usr/*"),
            ["usr/bin", "usr\\lib"] if normsep else ["usr/bin"],
        )
        self.assertEqual(
            filter(["usr/bin", "usr", "usr\\lib"], "usr\\*"),
            ["usr/bin", "usr\\lib"] if normsep else ["usr\\lib"],
        )


if __name__ == "__main__":
    unittest.main()



import typing
import io

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from . import mbuiltins

__mlib__ = "AbcLzma"


class _interface(ABC):
    import lzma as local_lzma

    @property
    @abstractmethod
    def LZMACompressor(self) -> local_lzma.LZMACompressor:
        ...

    @property
    @abstractmethod
    def LZMADecompressor(self) -> local_lzma.LZMADecompressor:
        ...

    @property
    @abstractmethod
    def LZMAFile(self) -> local_lzma.LZMAFile:
        ...

    @property
    @abstractmethod
    def LZMAError(self) -> local_lzma.LZMAError:
        ...

    @property
    @abstractmethod
    def _MODE_CLOSED(self):
        return self.local_lzma._MODE_CLOSED

    @property
    @abstractmethod
    def _MODE_READ(self):
        return self.local_lzma._MODE_READ

    @property
    @abstractmethod
    def _MODE_WRITE(self):
        return self.local_lzma._MODE_WRITE

    @property
    @abstractmethod
    def CHECK_NONE(self):
        return self.local_lzma.CHECK_NONE

    @property
    @abstractmethod
    def CHECK_CRC32(self):
        return self.local_lzma.CHECK_CRC32

    @property
    @abstractmethod
    def CHECK_CRC64(self):
        return self.local_lzma.CHECK_CRC64

    @property
    @abstractmethod
    def CHECK_SHA256(self):
        return self.local_lzma.CHECK_SHA256

    @property
    @abstractmethod
    def CHECK_ID_MAX(self):
        return self.local_lzma.CHECK_ID_MAX

    @property
    @abstractmethod
    def CHECK_UNKNOWN(self):
        return self.local_lzma.CHECK_UNKNOWN

    @property
    @abstractmethod
    def FILTER_LZMA1(self):
        return self.local_lzma.FILTER_LZMA1

    @property
    @abstractmethod
    def FILTER_LZMA2(self):
        return self.local_lzma.FILTER_LZMA2

    @property
    @abstractmethod
    def FILTER_DELTA(self):
        return self.local_lzma.FILTER_DELTA

    @property
    @abstractmethod
    def FILTER_X86(self):
        return self.local_lzma.FILTER_X86

    @property
    @abstractmethod
    def FILTER_IA64(self):
        return self.local_lzma.FILTER_IA64

    @property
    @abstractmethod
    def FILTER_ARM(self):
        return self.local_lzma.FILTER_ARM

    @property
    @abstractmethod
    def FILTER_ARMTHUMB(self):
        return self.local_lzma.FILTER_ARMTHUMB

    @property
    @abstractmethod
    def FILTER_POWERPC(self):
        return self.local_lzma.FILTER_POWERPC

    @property
    @abstractmethod
    def FILTER_SPARC(self):
        return self.local_lzma.FILTER_SPARC

    @property
    @abstractmethod
    def FORMAT_AUTO(self):
        return self.local_lzma.FORMAT_AUTO

    @property
    @abstractmethod
    def FORMAT_XZ(self):
        return self.local_lzma.FORMAT_XZ

    @property
    @abstractmethod
    def FORMAT_ALONE(self):
        return self.local_lzma.FORMAT_ALONE

    @property
    @abstractmethod
    def FORMAT_RAW(self):
        return self.local_lzma.FORMAT_RAW

    @property
    @abstractmethod
    def MF_HC3(self):
        return self.local_lzma.MF_HC3

    @property
    @abstractmethod
    def MF_HC4(self):
        return self.local_lzma.MF_HC4

    @property
    @abstractmethod
    def MF_BT2(self):
        return self.local_lzma.MF_BT2

    @property
    @abstractmethod
    def MF_BT3(self):
        return self.local_lzma.MF_BT3

    @property
    @abstractmethod
    def MF_BT4(self):
        return self.local_lzma.MF_BT4

    @property
    @abstractmethod
    def MODE_FAST(self):
        return self.local_lzma.MODE_FAST

    @property
    @abstractmethod
    def MODE_NORMAL(self):
        return self.local_lzma.MODE_NORMAL

    @property
    @abstractmethod
    def PRESET_DEFAULT(self):
        return self.local_lzma.PRESET_DEFAULT

    @property
    @abstractmethod
    def PRESET_EXTREME(self):
        return self.local_lzma.PRESET_EXTREME

    @abstractmethod
    def open(self, filename, mode="rb", *, format=None, check=-1, preset=None, filters=None, encoding=None, errors=None, newline=None) -> io.TextIOWrapper | local_lzma.LZMAFile:
        ...

    @abstractmethod
    def compress(self, data, format=local_lzma.FORMAT_XZ, check=-1, preset=None, filters=None) -> typing.Any:
        ...

    @abstractmethod
    def decompress(self, data, format=local_lzma.FORMAT_AUTO, memlimit=None, filters=None) -> typing.Any:
        ...

    @abstractmethod
    def is_check_supported(self, check) -> bool:
        ...

    @abstractmethod
    def _decode_filter_properties(self, filter, data) -> typing.Any:
        ...

    @abstractmethod
    def _encode_filter_properties(self, filter, data) -> typing.Any:
        ...


class AbcLzma(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mlzma"

    __all__ = [
        "CHECK_NONE",
        "CHECK_CRC32",
        "CHECK_CRC64",
        "CHECK_SHA256",
        "CHECK_ID_MAX",
        "CHECK_UNKNOWN",
        "FILTER_LZMA1",
        "FILTER_LZMA2",
        "FILTER_DELTA",
        "FILTER_X86",
        "FILTER_IA64",
        "FILTER_ARM",
        "FILTER_ARMTHUMB",
        "FILTER_POWERPC",
        "FILTER_SPARC",
        "FORMAT_AUTO",
        "FORMAT_XZ",
        "FORMAT_ALONE",
        "FORMAT_RAW",
        "MF_HC3",
        "MF_HC4",
        "MF_BT2",
        "MF_BT3",
        "MF_BT4",
        "MODE_FAST",
        "MODE_NORMAL",
        "PRESET_DEFAULT",
        "PRESET_EXTREME",
        "LZMACompressor",
        "LZMADecompressor",
        "LZMAFile",
        "LZMAError",
        "open",
        "compress",
        "decompress",
        "is_check_supported",
    ]

    __doc__ = _interface.local_lzma.__doc__

    __delegated__ = {_interface: _interface.local_lzma}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_lzma, self)

        self.__delegate__(_interface, mod)



# -*- coding: utf-8 -*-

"""
**Advanced shell**

This module provides advanced shell operations such as polling and predefined runner.

.. seealso:: :class:`myrrh.core.service.IShellService`
"""
import statistics
import threading
import weakref

from myrrh.utils import mtimer, mshlex

from myrrh.core.interfaces import IInStream, IProcess
from myrrh.core.services.system import AbcRuntime, InheritedPropertyClass
from myrrh.utils import mstring

__mlib__ = "AbcAdvSh"

_FAKE_PROC_START_PID = 90000
_fake_proc_count = 0


def _fakepid():
    global _fake_proc_count
    _fake_proc_count += 1
    return _fake_proc_count + _FAKE_PROC_START_PID


class AbcAdvSh(AbcRuntime):
    __frameworkpath__ = "msh.advsh"

    class _AdvShellExpired(Exception):
        def __init__(self, cmd, timeout, output=None, stderr=None):
            self.cmd = cmd
            self.timeout = timeout
            self.output = output
            self.stderr = stderr

        def __repr__(self):
            return "Command '%s' timed out after %s seconds and %s execution%s" % (
                self.cmd,
                self.timeout,
                self.cmd.ncalls,
                not self.cmd.ncalls and "" or "s",
            )

    class TTLExpired(_AdvShellExpired):
        pass

    class TimeoutExpired(_AdvShellExpired):
        pass

    class Abort(Exception):
        pass

    class AbortAndContinue(Abort):
        pass

    class execute(metaclass=InheritedPropertyClass(list)):  # type: ignore[misc]
        def __init__(
            self,
            cmd,
            *,
            path=None,
            encoding=None,
            errors=None,
            count=None,
            during=None,
            interval=None,
            ttl=None,
            timeout=None,
            poll=None,
            raiseonttl=None,
            raiseontimeout=True,
        ):
            self._cmd = cmd
            if isinstance(cmd, (str, bytes)):
                if path:
                    raise ValueError("bytes or string is not supported for cmd when path parameter is set")

                self._cast = encoding and mstring.cast(cmd, encoding=encoding, errors=errors) or self.myrrh_os.shcast(cmd)
                self._args = encoding and mstring.typebytes(encoding, errors)(cmd) or self.myrrh_os.shencode(cmd)
            else:
                if not path:
                    raise ValueError("only bytes or string is supported for cmd when path parameter is not set")
                self._cast = encoding and mstring.cast(cmd, encoding=encoding, errors=errors) or self.myrrh_os.fscast(cmd[0])
                self._args = [(encoding and mstring.typebytes(encoding, errors)(a) or self.myrrh_os.fsencode(a)) for a in cmd]

            self.__poll = poll
            self.__ttl = ttl
            self.__raiseonttl = raiseonttl
            self.__raiseontimeout = raiseontimeout

            eval_count = (not during and 1 or None) if count is None else count

            self.__duration_timer = mtimer.MTimer(timeout=during, step=during and interval or None, raise_on_expired=True)

            self.__global_timer = mtimer.MTimer(
                timeout=timeout,
                count=eval_count,
                step=not during and interval or None,
                raise_on_expired=True,
            )

            self.__ttl_timer = mtimer.MTimer(timeout=ttl, raise_on_expired=True)

            # evaluate cmd
            _, timeout = self._timeout
            self.__oneexecution = (not count and not timeout and not during) or (not poll and not count and eval_count == 1)

            if path:
                self._path = encoding and mstring.typebytes(path, encoding=encoding, errors=errors) or self.myrrh_os.fsencode(path)
            else:
                self._path = None

            if not self._path and timeout is not None:
                args = self.myrrh_os.getdefaultshellb(self._args)
                self._path = args[0]
                self._args = args
            elif path:
                self._args = mshlex.list2cmdlineb(self._args)

            if self.__oneexecution:
                self.values()

        @property
        def _timeout(self):
            # if timeout define and ttl define
            timeout = self.__poll
            timeout_cause = "polling"

            if self.__ttl_timer.timeout is not None:
                timeout = self.__ttl_timer.timeout if timeout is None else min(timeout, self.__ttl_timer.timeleft)
                timeout_cause = timeout_cause if (timeout is None or timeout < self.__ttl_timer.timeleft) else "killed"

            if self.__global_timer.timeout is not None:
                timeout = timeout is not None and min(timeout, self.__global_timer.timeleft) or self.__global_timer.timeleft
                timeout_cause = timeout_cause if (timeout is None or timeout < self.__global_timer.timeleft) else "timeout"

            return timeout_cause, timeout

        def __repr__(self):
            cmd = self._cast(self._path + b" ") or self._cast(b"")
            cmd += self._cast(b" ".join(self._args) if isinstance(self._args, (list, tuple)) else self._args)

            if self.__oneexecution:
                return "<%s> %s" % (cmd, self[0].state)
            return "<%s>*%d in %ds" % (cmd, len(self), self.time)

        def _run(self):
            with self.__global_timer:
                _, timeout = self._timeout

                while True:
                    if not self._path and timeout is None:
                        proc = _AdvshellFakeProc(self, self._args)
                    else:
                        proc = _AdvshellProc(self, self._args, self._path)

                    try:
                        with self.__ttl_timer:
                            try:
                                with _ExecutionInformation(proc=proc, cast=self._cast) as exe:
                                    proc.exec()

                                    self.append(exe)
                                    cause = "polling"
                                    while cause == "polling":
                                        cause, timeout = self._timeout

                                        try:
                                            exe.communicate(timeout=timeout)
                                            cause = "exited"

                                        except mtimer.MTimer.DelayExpired:
                                            if cause == "polling":
                                                yield self[-1]

                                        except Exception:
                                            cause = "exception"

                                yield self[-1]

                                if cause == "killed" and self.__raiseonttl:
                                    raise AdvSh.TTLExpired(
                                        cmd=self,
                                        timeout=self.__ttl,
                                        output=exe.out,
                                        stderr=exe.err,
                                    ) from None

                                self.__duration_timer.idle
                                self.__global_timer.idle

                            except GeneratorExit:
                                if exe.state != "closed":
                                    exe.state = "aborted"
                                break
                            except mtimer.MTimer.CountExpired:
                                break
                            except mtimer.MTimer.DelayExpired as e:
                                if self.__raiseontimeout and not e.timer == self.__duration_timer:
                                    raise AdvSh.TimeoutExpired(
                                        cmd=self,
                                        timeout=self.__global_timer.duration,
                                        output=exe.out,
                                        stderr=exe.err,
                                    ) from None
                                break

                    finally:
                        proc.close()

        def __iter__(self):
            # execution ended
            if self.__global_timer.endtime:
                for exe in self.calls:
                    if self.__oneexecution:
                        yield from exe
                    else:
                        yield exe
            else:
                for exe in self._run():
                    yield exe

        def values(self):
            if not self.__global_timer.endtime:
                for _ in self:
                    pass
            return self

        def iter(self):
            yield self.__iter__()

        def itervalues(self):
            for exe in self.calls:
                yield exe

        # general execution information
        @property
        def cmd(self):
            return self._cast(self._cmd)

        @property
        def ncalls(self):
            return len(self)

        @property
        def calls(self):
            return list.__iter__(self)  # 3.5 fixes

        @property
        def time(self):
            return self.__global_timer.duration

        @property
        def mtime(self):
            try:
                return statistics.mean(map(lambda x: x.time, self.calls))
            except statistics.StatisticsError:
                return 0

        @property
        def vtime(self):
            try:
                return statistics.variance(map(lambda x: x.time, self.calls))
            except statistics.StatisticsError:
                return None


class _AdvshellFakeProc:
    def __init__(self, system: AbcAdvSh, cmd: bytes):
        self.system = weakref.proxy(system)
        self.cmd = cmd
        self._pid = None
        self.closed = False

    def exec(self):
        if self._pid is not None:
            return

        o, e, r = self.system.myrrh_os.shell.execute(self.cmd)
        self._out = o
        self._err = e
        self._rval = r

        self._pid = _fakepid()
        self.closed = True

    def error(self, nbytes=None, timeout=None):
        nbytes = len(self._err) if nbytes is None else min(len(self._err), nbytes)
        res = self._err[:nbytes]
        self._err = self._err[nbytes:]
        return res or None

    def output(self, nbytes=None, timeout=None):
        nbytes = len(self._out) if nbytes is None else min(len(self._out), nbytes)
        res = self._out[:nbytes]
        self._out = self._out[nbytes:]
        return res or None

    @property
    def exit_status(self):
        return self._rval

    @property
    def is_running(self):
        return False

    def terminate(self):
        pass

    @property
    def pid(self):
        return self._pid

    def close(self):
        pass


class _AdvshellProc:
    def __init__(self, system: AbcAdvSh, args: list[bytes], path: bytes):
        self.system = weakref.proxy(system)
        self.args = args
        self.path = path
        self._pid = None
        self.closed = False

    def exec(self):
        if self._pid is not None:
            return

        outr = outw = errr = errw = None
        self._out = self._err = None

        try:
            outr, outw = self.system.myrrh_syscall.open_pipe()
            errr, errw = self.system.myrrh_syscall.open_pipe()

            hproc = self.system.myrrh_syscall.open_process(self.path, self.args, stdout=outw, stderr=errw)

        except:  # noqa: E722
            if outw is not None:
                self.system.myrrh_syscall.close(outr)  # type: ignore[arg-type]
            if errw is not None:
                self.system.myrrh_syscall.close(errr)  # type: ignore[arg-type]
            raise

        finally:
            if outw is not None:
                self.system.myrrh_syscall.close(outw)
            if errw is not None:
                self.system.myrrh_syscall.close(errw)

        self._proc: IProcess = self.system.myrrh_syscall.gethandle(hproc)  # type: ignore
        self._out: IInStream = self.system.myrrh_syscall.gethandle(outr)  # type: ignore
        self._err: IInStream = self.system.myrrh_syscall.gethandle(errr)  # type: ignore

        self._exit_status = None
        self._pid = self._proc.pid

    def error(self, nbytes=None, timeout=None):
        if self._err is None:
            return b""

        if not self.closed and self._err:
            return self._err.read(nbytes, extras={"timeout": timeout})

        raise BrokenPipeError("operation on closed process is not available")

    def output(self, nbytes=None, timeout=None):
        if self._out is None:
            return b""

        if not self.closed:
            return self._out.read(nbytes, extras={"timeout": timeout})

        raise BrokenPipeError("operation on closed process is not available")

    def close(self):
        if not self.closed and self._pid:
            self._proc.close()
            self._out.close()
            self._err.close()
        self.closed = True

    @property
    def exit_status(self):
        if self._exit_status is None and self.closed:
            raise RuntimeError(f"the return value of the process {self._pid} is lost")

        if self._exit_status is None:
            self._exit_status = self._proc.exit_status

        return self._exit_status

    @property
    def is_running(self):
        return self.exit_status is None

    def terminate(self):
        if not self.closed and self._pid:
            self._proc.terminate()
            self._exit_status = self.system.myrrh_syscall.wait(self._proc)

    @property
    def pid(self):
        return self._pid


class _ExecutionInformation:
    MAX_BUFFER_SIZE = 50000

    def __init__(self, proc, cast, **kwargs):
        self._proc = proc
        self._cast = cast
        self._timer = mtimer.MTimer()
        self._state = None

        self._buflock = threading.RLock()
        self._out = bytearray(self.MAX_BUFFER_SIZE)
        self._err = bytearray(self.MAX_BUFFER_SIZE)

        self._closed = False

        for k, v in kwargs:
            setattr(self, k, v)

    def __iter__(self):
        yield from (self.out, self.err, self.rval)

    def __del__(self):
        if not self._closed:
            self._proc.terminate()

    def __getitem__(self, item):
        if hasattr(self, item):
            return getattr(self, item)
        raise KeyError(item)

    def __enter__(self):
        self._timer.reset()
        return self

    def __exit__(self, *_a, **_k):
        self.close()
        self._timer.stop()

    def communicate(self, timeout=None):
        # wait en of proc
        com_timer = mtimer.MBurstTimer(timeout=timeout, raise_on_expired=True)
        self._out.clear()
        self._err.clear()

        while True:
            if self._proc.exit_status is not None:
                break

            try:
                self._out.extend(self._proc.output(timeout=com_timer.timeleft) or b"")
            except TimeoutError:
                ...

            try:
                self._err.extend(self._proc.error(timeout=com_timer.timeleft) or b"")
            except TimeoutError:
                ...
            com_timer.idle

    def close(self):
        if not self._closed:
            self.proc.terminate()
            self._out.extend(self._proc.output() or b"")
            self._err.extend(self._proc.error() or b"")
            self.proc.close()
            self._closed = True

    @property
    def closed(self):
        return self._closed

    @property
    def proc(self):
        return self._proc

    @property
    def time(self):
        return self._timer.duration

    @property
    def starttime(self):
        return self._timer.starttime

    @property
    def endtime(self):
        return self._timer.endtime

    @property
    def state(self):
        state, _, _ = self._state or (None, None, None)
        if self._proc.exit_status is None:
            return "-".join(filter(None, ["running", state]))

        return "-".join(filter(None, ["terminated", state]))

    @state.setter
    def state(self, state):
        self._state = (
            state,
            self._out and len(self._out) or 0,
            self._err and len(self._err) or 0,
        )

    @property
    def out(self):
        return self._cast(self._out)

    @property
    def err(self):
        return self._cast(self._err)

    @property
    def rval(self):
        return self._proc.exit_status


AdvSh = AbcAdvSh

Abort = AdvSh.Abort
AbortAndContinue = AdvSh.AbortAndContinue
TimeoutExpired = AdvSh.TimeoutExpired
TTLExpired = AdvSh.TTLExpired



import typing
import click

import bmy


from myrrh.tools.myrrhc_ext import myrrhc_cmds, types, cmd
from myrrh.provider.registry import ProviderRegistry
from myrrh.warehouse.registry import ItemRegistry


@myrrhc_cmds.group()
def provider():
    "Provider management"
    pass


@provider.command(name="list")
def list_():
    "List all installed providers"
    for name in sorted(ProviderRegistry().providers, key=lambda n: n[0]):
        cmd.info("%s\t\t%s" % (name, ProviderRegistry().get_version(name)))


@provider.group()
def new():
    ...


for pname in sorted(ProviderRegistry().providers, key=lambda n: n[0]):

    def _(eid, pname=pname, **kw):
        eid = bmy.new(f"**/{pname}", eid=eid, **kw)
        bmy.select(eid)

    settings = ItemRegistry().provider_settings.get(pname)

    if settings:
        for field_name, field_info in settings.model_fields.items():
            if field_name != "name":
                types_ = [
                    a
                    for a in filter(
                        lambda t: not issubclass(t, None.__class__),
                        typing.get_args(field_info.annotation),
                    )
                ]

                if not types_:
                    type_ = field_info.annotation

                elif len(types_) == 1:
                    type_ = types_[0]

                else:

                    class Type_(click.ParamType):
                        name = f"{field_info.annotation}"
                        types_ = types_

                        def convert(self, value, param, ctx):
                            for t in self.types_:
                                try:
                                    return t(value)
                                except (TypeError, ValueError):
                                    pass

                            raise ValueError(f"invalid value for '{param}': '{value}' is not valid for '{self.name}' types")

                    type_ = Type_()
                nargs = typing.get_origin(field_info.annotation) and issubclass(typing.get_origin(field_info.annotation), (tuple, list))
                _ = cmd.option(
                    f"--{field_name}",
                    required=field_info.is_required(),
                    type=type_,
                    is_flag=type_ is bool,
                    multiple=nargs,
                    default=field_info.get_default(),
                    show_default=True,
                    help=field_info.description,
                )(_)

    _ = cmd.argument("eid", type=types.STRING)(_)
    new.command(name=pname)(_)



from myrrh.tools.myrrhc_ext import myrrhc_cmds, cmd
from myrrh.provider.registry import ProviderRegisry


@myrrhc_cmds.group()
def provider():
    "Provider management"
    pass


@provider.command()
def list():
    "List all installed providers"
    for name in sorted(ProviderRegisry().providers, key=lambda n: n[0]):
        cmd.info("%s\t\t%s" % (name, ProviderRegisry().get_version(name)))



# -*- coding: utf-8 -*-

"""
**like c getch implementation **

.. note:: this module should not be used outside Myrrh library.

This module implements a delegation solution for abstract method delegation

-----------

"""
import threading

import time
from collections import deque
from prompt_toolkit.input.defaults import create_input
from prompt_toolkit.key_binding import KeyPress

import asyncio

__all__ = ["getchIt", "getch", "getche"]


class _getChPromptToolkit:
    _lock = threading.RLock()
    _keys: deque[KeyPress] = deque()
    _keyevt = threading.Condition(_lock)
    _reader: threading.Thread | None = None
    _emergency = False

    @classmethod
    def abort(cls):
        with cls._lock:
            if not cls._reader:
                return

            cls._reader = None
            cls._keyevt.wait()

    @classmethod
    def _read_runner(cls):
        input = create_input()

        try:
            asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        with input.raw_mode():
            with input.detach():
                no_emergency = 3
                while True:
                    with cls._lock:
                        keys = input.read_keys()
                        # emergency escape
                        for k in keys:
                            if k.key == "c-c":
                                no_emergency -= 1
                            else:
                                no_emergency = 3

                        if not no_emergency:
                            keys.append(KeyPress("!", data="c-c"))

                        if keys:
                            cls._keys.extend(keys)

                        if keys or not cls._reader:
                            cls._keyevt.notifyAll()
                        if not cls._reader:
                            break
                    if not keys:
                        time.sleep(0.05)

    @classmethod
    def read(cls):
        with cls._lock:
            while True:
                if cls._keys:
                    return cls._keys.popleft()
                elif cls._reader:
                    cls._keyevt.wait()
                else:
                    return None

    @classmethod
    def unset(cls):
        with cls._lock:
            cls.abort()
            cls._keys = None

    @classmethod
    def set(cls, echo=False):
        with cls._lock:
            if not cls._reader:
                cls._keys = deque()
                cls._reader = threading.Thread(name="getch", target=cls._read_runner, daemon=True)
                cls._reader.start()


impl = _getChPromptToolkit


class getchIt(object):
    def __enter__(self):
        impl.set()
        return self

    def __exit__(self, *_err):
        impl.unset()

    def __iter__(self):
        return self

    def __next__(self):
        result = ""
        while not result:
            result = impl.read()
            if result is None:
                raise StopIteration
            if result.key == "!" and result.data == "c-c":
                raise KeyboardInterrupt

        return result.data


class getcheIt(getchIt):
    def __enter__(self):
        try:
            impl.set(echo=True)
        except RuntimeError:
            self.leave = False
        return self

    def __exit__(self, *_err):
        if self.leave:
            impl.unset()


def getch():
    with getchIt() as it:
        try:
            ch = next(it)
            return b"" if ch is None else ch
        except:  # noqa: E722
            return b""


def getche():
    with getcheIt() as it:
        ch = next(it)
        return b"" if ch is None else ch


def abort():
    impl.abort()



import inspect

from abc import ABCMeta, abstractmethod, ABC
import typing

__all__ = (
    "ABCDelegationMeta",
    "NoneDelegation",
    "ABCDelegation",
    "ABCMeta",
    "abstractmethod",
    "ABC",
    "DelegateProperty",
)


class DelegateProperty:
    def __init__(self, cls, name):
        self.cls = cls
        self.name = name

    def __get__(self, obj, type):
        if obj is None:
            return self

        try:
            return getattr(obj._delegate_, self.name)
        except AttributeError:
            pass

    def __set__(self, obj, value):
        setattr(obj._delegate_, self.name, value)

    def __delete__(self, obj):
        delattr(obj._delegate_, self.name)


class Delegation:
    __delegation_ref__: dict

    __delagation_owner__: typing.Any = None

    def __init__(self):
        super().__setattr__("__delegation_ref__", dict())
        super().__setattr__("__delagation_owner__", None)

    def __set_name__(self, owner, name):
        super().__setattr__("__delagation_owner__", owner)

    def __getattr__(self, name: str):
        try:
            _, ref, _getattr_ = self.__delegation_ref__[name]
            if ref is not None:
                return _getattr_(ref, name)
        except KeyError:
            pass

        raise AttributeError(f"{self.__delagation_owner__.__class__.__name__} delegation has no attribute {name}")

    def __setattr__(self, name: str, value):
        try:
            _, ref, _ = self.__delegation_ref__[name]
            if ref is not None:
                return setattr(ref, name, value)
        except KeyError:
            pass

        raise AttributeError(f"{self.__delagation_owner__.__class__.__name__} delegation has no attribute {name}")

    def __delattr__(self, name: str):
        try:
            _, ref, _ = self.__delegation_ref__[name]
            if ref is not None:
                return delattr(ref, name)
        except KeyError:
            pass

        raise AttributeError(f"{self.__delagation_owner__.__class__.__name__} delegation has no attribute {name}")


class ABCDelegationMeta(ABCMeta):
    def __prepare__(name, bases):
        dct = {"__delegate__": ABCDelegationMeta.__delegate__, "_delegate_": None}

        return dct

    def __new__(mcls, name, bases, dct):
        delegated = dct.get("__delegated__") or dict()
        delegated_attrs = dct.get("__delegated_attrs__") or set()

        if not isinstance(delegated, dict):
            delegated = dict.fromkeys(delegated, None)

        for b in bases:
            if hasattr(b, "__delegated__"):
                delegated.update(b.__delegated__)
            if hasattr(b, "__delegated_attrs__"):
                delegated_attrs.update(b.__delegated_attrs__)

        dct["__delegated__"] = delegated
        dct["__delegated_attrs__"] = delegated_attrs

        inherited_dct = set(m for b in bases for m, _ in inspect.getmembers_static(b) if m not in (getattr(b, "__abstractmethods__", None) or list()))

        for delgcls in delegated:
            if not hasattr(delgcls, "__abstractmethods__"):
                raise TypeError(f"{delgcls.__name__} invalid type: delegated class type must be abc.ABCMeta")

            for method in delgcls.__abstractmethods__:
                dct["__delegated_attrs__"].add(method)
                if method not in dct and method not in inherited_dct:
                    dct[method] = DelegateProperty(delgcls, method)

        return super().__new__(mcls, name, bases, dct)

    def __call__(cls, *a, **kwa):
        inst = cls.__new__(cls, *a, **kwa)

        inst._delegate_ = Delegation()
        inst._delegate_.__set_name__(inst, "_delegate_")

        for delgcls, default in cls.__delegated__.items():
            if default:
                for m in delgcls.__abstractmethods__:
                    inst._delegate_.__delegation_ref__[m] = (delgcls, default, getattr)

        if isinstance(inst, cls):
            inst.__init__(*a, **kwa)

        return inst

    def __delegate__(self, cls, obj, *, checktype=None):
        if checktype is None:
            checktype = getattr(self, "__delegate_check_type__", True)

        delegate_all = getattr(cls, "__delegate_all__", None) or [cls]

        if checktype and delegate_all and not all(isinstance(obj, cls_) for cls_ in delegate_all):
            raise Exception("invalid delegation for %s : %s invalid type" % (cls.__name__, obj.__class__.__name__))

        if self is obj:
            if not hasattr(obj, "__get_delegate__"):
                raise Exception("invalid delegation for %s : %s need  __get_delegate__ method" % (cls.__name__, obj.__class__.__name__))

            for m in cls.__abstractmethods__:
                self._delegate_.__delegation_ref__[m] = (
                    cls,
                    obj,
                    obj.__class__.__get_delegate__,
                )
        else:
            for m in cls.__abstractmethods__:
                delegated = getattr(obj, "_delegate_", None)
                getter = getattr

                if delegated:
                    cls_, obj_, getter_ = delegated.__delegation_ref__.get(m) or (
                        None,
                        None,
                        None,
                    )
                    if cls_ is cls and isinstance(getter_(obj_.__class__, m, None), DelegateProperty):
                        cls, obj, getter = cls_, obj_, getter_

                self._delegate_.__delegation_ref__[m] = cls, obj, getter


class ABCDelegation(metaclass=ABCDelegationMeta):
    __delegated__: tuple[typing.Any, ...] | typing.Dict[typing.Type[ABC], typing.Any] | None = None
    __delegate_all__: tuple[typing.Any, ...]
    __delegate_check_type__: bool

    @property
    @abstractmethod
    def _delegate_(self) -> typing.Any:
        ...

    @abstractmethod
    def __delegate__(self, t: typing.Type[ABC], o: object) -> None:
        ...


def NoneDelegation(name, delegation_class):
    def __new__(cls):
        if cls.__inst__:
            return cls.__inst__
        return object.__new__(cls)

    def __init__(self):
        self.__delegate__(delegation_class, self)

    def __bool__(self):
        return False

    def __get_delegate__(self, name):
        raise AttributeError(f"{delegation_class} is not provided")

    d = ABCDelegationMeta.__prepare__(name, (delegation_class,))
    d.update(
        {
            "__delegated__": (delegation_class,),
            "__inst__": None,
            "__new__": __new__,
            "__init__": __init__,
            "__bool__": __bool__,
            "__get_delegate__": __get_delegate__,
        }
    )

    cls = ABCDelegationMeta(name, (delegation_class,), d)

    inst = cls()
    cls.__inst__ = inst

    return inst



import typing

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from myrrh.framework.mpython import mbuiltins

__mlib__ = "AbcScriptHelper"


class _interface(ABC):
    import test.support.script_helper as local_script_helper

    @abstractmethod
    def assert_python_ok(self, *args, **env_vars) -> typing.Any:
        ...

    @abstractmethod
    def assert_python_failure(self, *args, **env_vars) -> typing.Any:
        ...

    @abstractmethod
    def run_python_until_end(self, *args, **env_vars) -> typing.Any:
        ...


class AbcScriptHelper(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mtest.msupport.mscript_helper"

    __delegated__ = {_interface: _interface.local_script_helper}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_script_helper, self)
        self.__delegate__(_interface, mod)



import errno
import array
import os

from myrrh.core.services.system import MIOException, AbcRuntime

from . import mimportlib

__mlib__ = "AbcOsFile"


class AbcOsFile(AbcRuntime):
    __frameworkpath__ = "mpython._mosfile"

    __all__ = [
        "SEEK_SET",
        "SEEK_CUR",
        "SEEK_END",
        "O_RDONLY",
        "O_WRONLY",
        "O_BINARY" "O_RDWR",
        "O_APPEND",
        "O_CREAT",
        "O_TRUNC",
        "O_EXCL",
        "fdopen",
        "close",
        "closerange",
        "dup",
        "dup2",
        "fchmod",
        "fchown",
        "fdatasync",
        "fstat",
        "fsync",
        "ftruncate",
        "isatty",
        "lseek",
        "open",
        "read",
        "write",
        "set_inheritable",
        "get_inheritable",
    ]

    _BUFFER_SZ = 8192

    from os import O_WRONLY, O_RDWR, O_APPEND, O_CREAT, O_TRUNC, O_EXCL, O_RDONLY

    try:
        from os import O_BINARY
    except ImportError:
        pass

    O_TEMPORARY = 64

    osfs = mimportlib.module_property("_osfs")

    def _fs_stream(self, fd):
        return self.myrrh_syscall.gethandle(fd, True)

    def close(self, fd):
        return self.myrrh_syscall.close(fd)

    def closerange(self, fd_low, fd_high):
        for fd in range(fd_low, fd_high):
            try:
                self.close(fd)
            except OSError:
                pass

    def dup(self, fd):
        return self.myrrh_syscall.dup(fd)

    def dup2(self, fd, fd2, inheritable=True):
        return self.myrrh_syscall.dup2(fd, fd2)

    def fchmod(self, fd, mode):
        return self.osfs.chmod(self._fs_stream(fd).path, mode)

    def fchown(self, fd, uid, gid):
        return self.osfs.chown(self._fs_stream(fd).path, uid, gid)

    def fdatasync(self, fd):
        return self.fsync(fd)

    def fstat(self, fd):
        handle = self._fs_stream(fd)
        try:
            stat = handle.stat()
        except NotImplementedError:
            stat = self.osfs.stat(handle.path)

        stat = os.stat_result(
            (
                stat.st_mode,
                stat.st_ino,
                stat.st_dev,
                stat.st_nlink,
                stat.st_uid,
                stat.st_gid,
                stat.st_size,
                stat.st_atime,
                stat.st_mtime,
                stat.st_ctime,
            )
            + (0,) * (os.stat_result.n_fields - 10)
        )
        return stat

    def fsync(self, fd):
        try:
            return self._fs_stream(fd).sync()
        except (IOError, OSError):
            raise
        except Exception as e:
            MIOException(self, exc=e).raised()

    def ftruncate(self, fd, length):
        try:
            return self._fs_stream(fd).truncate(length)
        except (IOError, OSError):
            raise
        except Exception as e:
            MIOException(self, exc=e).raised()

    def isatty(self, fd):
        return False

    def lseek(self, fd, pos, how):
        try:
            return self._fs_stream(fd).seek(pos, how)

        except (IOError, OSError, OverflowError, ValueError):
            raise
        except Exception as e:
            MIOException(self, exc=e).raised()

    def open(self, path, flags, mode=0o777, *, dir_fd=None) -> int:
        file = self.myrrh_os.p(path, dir_fd=dir_fd)

        if b"\0" in file:
            raise ValueError("embedded null character")

        mode = mode & ~self.osfs._umask
        abspath = self.myrrh_os.getpathb(file)

        if b"\x00" in abspath:
            raise ValueError("embedded null byte")

        # check file
        if (self.O_CREAT & flags) == 0:
            if not self.osfs.exists(abspath):
                MIOException(self, errno.ENOENT, filename=path).raised()
        else:
            if (self.O_EXCL & flags) != 0:
                if self.osfs.exists(abspath):
                    MIOException(self, errno.EEXIST, filename=path).raised()

        wiring = self.myrrh_syscall.Wiring.OFF
        if flags & self.O_RDWR:
            wiring |= self.myrrh_syscall.Wiring.INOUT
        if flags & self.O_WRONLY:
            wiring |= self.myrrh_syscall.Wiring.OUT
        else:
            wiring |= self.myrrh_syscall.Wiring.IN

        if not flags & self.O_APPEND:
            wiring |= self.myrrh_syscall.Wiring.RESET

        if flags & self.O_CREAT:
            wiring |= self.myrrh_syscall.Wiring.CREATE

        try:
            hint = self.myrrh_syscall.open_file(abspath, wiring, extras={"mode": mode, "flags": flags})
        except (IOError, OSError) as e:
            MIOException(self, errno=e.errno, filename=path, exc=e).raised()
        except Exception as e:
            MIOException(self, exc=e, filename=path).raised()

        return hint

    def read(self, fd, n):
        try:
            return self._fs_stream(fd).read(n)
        except (IOError, OSError):
            raise
        except Exception as e:
            MIOException(self, exc=e).raised()

    def write(self, fd, str):
        if isinstance(str, memoryview):
            str = str.tobytes()

        if isinstance(str, (bytearray, array.array)):
            str = bytes(str)

        if not isinstance(str, bytes):
            raise TypeError("a bytes-like object is required, not 'str'")

        try:
            return self._fs_stream(fd).write(str)
        except (IOError, OSError):
            raise
        except Exception as e:
            MIOException(self, exc=e).raised()

    # stub
    def set_inheritable(self, fd, inheritable):
        self._fs_stream(fd)

    def get_inheritable(self, fd):
        self._fs_stream(fd)
        return False



# flake8: noqa: F401
from ._entity import *
from ._services import *

from ..services import cfg_init

cfg_init("eid_prefix", "entity", section="myrrh.core.entity")
cfg_init("validate_service_args", True, section="myrrh.core.entity")



from collections import namedtuple

from myrrh.core.services.system import ExecutionFailureCauseRVal

from myrrh.framework.mpython.msys import AbcSys

__mlib__ = "Sys"


class Sys(AbcSys):
    __executable: bytes | None = None

    __all__ = AbcSys.__all__ + [
        "getwindowsversion",
    ]

    def _enablelegacywindowsfsencoding(self):
        self.myrrh_os.fsencoding = "mbcs"

    @property
    def _platform(self):
        return "win32"

    @property
    def _executable(self):
        if self.__executable is None:
            self.__executable, err, rval = self.myrrh_os.cmdb(b"%(where)s python")
            self.__executable = self.__executable.split(b"\r\n")[0] if self.__executable else self.__executable
            ExecutionFailureCauseRVal(self, err, rval, 0).check()

        return self.__executable

    def getwindowsversion(self):
        major, minor, build = (int(v) for v in self.myrrh_os._wininfo["Version"].split("."))
        version = namedtuple(
            "sys_getwindowsversion",
            "major minor build platform service_pack service_pack_major service_pack_minor suite_mast product_type platform_version",
        )

        return version(
            major,
            minor,
            build,
            2,
            self.myrrh_os._wininfo["CSDVersion"],
            self.myrrh_os._wininfo["ServicePackMajorVersion"],
            self.myrrh_os._wininfo["ServicePackMinorVersion"],
            self.myrrh_os._wininfo["OSProductSuite"],
            self.myrrh_os._wininfo["ProductType"],
            (major, minor, build),
        )
